"""
Utilities functions
"""

# Standard library modules.

# Third party modules.

# Local modules.

# Globals and constants variables.

c = 299792458
h_eVs = 4.13566733e-15


def energy_to_wavelength_m(energy_eV):
    return h_eVs * c / energy_eV


def wavelength_to_energy_eV(wavelength_m):
    return h_eVs * c / wavelength_m



#!/usr/bin/env python
""" """

# Standard library modules.
import os

# Third party modules.
import pytest

# Local modules.

# Globals and constants variables.


@pytest.fixture
def testdatadir():
    return os.path.join(os.path.dirname(__file__), "testdata")



"""
pyxray - Definitions and properties of X-ray transitions
"""

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from pyxray.base import NotFound
from pyxray.descriptor import *
from pyxray.data import *
from pyxray.composition import *



"""
Base database
"""

# Standard library modules.
import abc
import sys
import operator

# Third party modules.
import tabulate

# Local modules.
from pyxray.cbook import formatdoc
import pyxray.descriptor as descriptor

# Globals and constants variables.


class NotFound(Exception):
    pass


_docextras = {
    "element": """:arg element: either
            * :class:`Element <pyxray.descriptor.Element>` object
            * atomic number
            * symbol (case insensitive)
            * name (in any language, case insensitive)
            * object with attribute :attr:`atomic_number` or :attr:`z`""",
    "atomic_shell": """:arg atomic_shell: either
            * :class:`AtomicShell <pyxray.descriptor.AtomicShell>` object
            * principal quantum number
            * any notation (case insensitive)""",
    "atomic_subshell": """:arg atomic_subshell: either
            * :class:`AtomicSubshell <pyxray.descriptor.AtomicSubshell>` object
            * a :class:`tuple` of principal quantum number,
            azimuthal quantum number, and total angular momentum_nominator
            * any notation (case insensitive)""",
    "xray_transition": """:arg xray_transition: either
            * :class:`XrayTransition <pyxray.descriptor.XrayTransition>` object
            * a :class:`tuple` of source and destination subshells
            * any notation (case insensitive)""",
    "language": """:arg language: language code (e.g. ``en``, ``fr``, ``de``)""",
    "notation": """:arg notation: name of a notation (case insensitive),
                ``iupac``, ``siegbahn`` and ``orbital`` are usually supported""",
    "encoding": """:arg encoding: type of encoding, either
                ``ascii``, ``utf16``, ``html`` or ``latex``""",
    "reference": """:arg reference: reference to use to retrieve this value, either
            * :class:`Reference <pyxray.descriptor.Reference>` object
            * BibTeX key of a reference
            * ``None``, the default reference will be used or the first reference found""",
    "exception": """:raise NotFound:""",
}


class _DatabaseMixin(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_preferred_references(self):
        """
        Return the BibTeX keys of the preferred references.

        If no reference is specified when calling a method, the value for the first preferred reference is returned.
        If no preferred reference, the first value is returned.

        :return: preferred references
        :rtype: :class:`tuple`
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def add_preferred_reference(self, reference):
        """
        Adds a preferred reference.

        :arg reference: :class:`Reference` or its BibTeX key

        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clear_preferred_references(self):
        """
        Clear all added preferred references.
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element(self, element):  # pragma: no cover
        """
        Returns element descriptor.

        {element}

        :return: element descriptor
        :rtype: :class:`Element`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_atomic_number(self, element):  # pragma: no cover
        """
        Returns atomic number of an element.

        {element}

        :return: atomic number
        :rtype: :class:`int`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_symbol(self, element, reference=None):  # pragma: no cover
        """
        Returns symbol of an element.

        {element}
        {reference}

        :return: symbol
        :rtype: :class:`str`
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_name(self, element, language="en", reference=None):  # pragma: no cover
        """
        Returns full name of an element, in the language specified.

        {element}
        {language}
        {reference}

        :return: name
        :rtype: :class:`str`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_atomic_weight(self, element, reference=None):  # pragma: no cover
        """
        Returns atomic weight of an element.
        The atomic weight is defined by the CIAAW as it is the ratio of
        the average atomic mass of an element over 1/12 of the mass of the
        carbon-12 atom.

        {element}
        {reference}

        :return: atomic weight
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_mass_density_kg_per_m3(
        self, element, reference=None
    ):  # pragma: no cover
        """
        Returns mass density (in kg/m3) of an element.

        {element}
        {reference}

        :return: mass density (in kg/m3)
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @formatdoc(**_docextras)
    def element_mass_density_g_per_cm3(
        self, element, reference=None
    ):  # pragma: no cover
        """
        Returns mass density (in g/cm3) of an element.

        {element}
        {reference}

        :return: mass density (in g/cm3)
        :rtype: :class:`float`
        {exception}
        """
        return self.element_mass_density_kg_per_m3(element, reference) / 1e3

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_xray_transition(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        """
        Returns X-ray transition descriptor if x-ray transition has a
        probability greater than 0 for that element.

        {element}
        {xray_transition}
        {reference}

        :return: X-ray transition descriptor
        :rtype: :class:`XrayTransition`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def element_xray_transitions(
        self, element, xray_transition=None, reference=None
    ):  # pragma: no cover
        """
        Returns all x-ray transitions which have a probability greater than 0 for that element.
        If *xray_transition* is not ``None``, only the x-ray transitions matching this x-ray transition are returned.

        {element}
        {xray_transition}
        {reference}

        :return: X-ray transitions
        :rtype: :class:`tuple` of :class:`XrayTransition`
        {exception}
        """
        raise NotImplementedError

    @formatdoc(**_docextras)
    def print_element_xray_transitions(
        self, element, file=sys.stdout, tabulate_kwargs=None
    ):
        """
        Prints all x-ray transitions for an element, with their different
        notations and energy.

        {element}

        :arg file: file for output, default to standard out
        """
        header = ["IUPAC", "Siegbahn", "Energy (eV)", "Probability", "Relative weight"]

        rows = []
        for xray_transition in self.element_xray_transitions(element):
            try:
                iupac = self.xray_transition_notation(xray_transition, "iupac")
            except:
                iupac = ""

            try:
                siegbahn = self.xray_transition_notation(xray_transition, "siegbahn")
            except:
                siegbahn = ""

            try:
                energy_eV = self.xray_transition_energy_eV(element, xray_transition)
            except:
                energy_eV = ""

            try:
                probability = self.xray_transition_probability(element, xray_transition)
            except:
                probability = ""

            try:
                relative_weight = self.xray_transition_relative_weight(
                    element, xray_transition
                )
            except:
                relative_weight = ""

            rows.append([iupac, siegbahn, energy_eV, probability, relative_weight])

        rows.sort(key=operator.itemgetter(2))

        if tabulate_kwargs is None:
            tabulate_kwargs = {}
        file.write(tabulate.tabulate(rows, header, **tabulate_kwargs))

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_shell(self, atomic_shell):  # pragma: no cover
        """
        Returns atomic shell descriptor.

        {atomic_shell}

        :return: atomic shell descriptor
        :rtype: :class:`AtomicShell`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_shell_notation(
        self, atomic_shell, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        """
        Returns notation of an atomic shell.

        {atomic_shell}
        {notation}
        {encoding}
        {reference}

        :return: notation
        :rtype: :class:`str`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell(self, atomic_subshell):  # pragma: no cover
        """
        Returns atomic subshell descriptor.

        {atomic_subshell}

        :return: atomic subshell descriptor
        :rtype: :class:`AtomicSubshell`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell_notation(
        self, atomic_subshell, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        """
        Returns notation of an atomic subshell.

        {atomic_subshell}
        {notation}
        {encoding}
        {reference}

        :return: notation
        :rtype: :class:`str`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell_binding_energy_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        """
        Returns binding energy of an element and atomic subshell (in eV).

        {element}
        {atomic_subshell}
        {reference}

        :return: binding energy (in eV)
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell_radiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        """
        Returns radiative width of an element and atomic subshell (in eV).

        {element}
        {atomic_subshell}
        {reference}

        :return: radiative width (in eV)
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell_nonradiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        """
        Returns nonradiative width of an element and atomic subshell (in eV).

        {element}
        {atomic_subshell}
        {reference}

        :return: nonradiative width (in eV)
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def atomic_subshell_occupancy(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        """
        Returns occupancy of an element and atomic subshell.

        {element}
        {atomic_subshell}
        {reference}

        :return: occupancy
        :rtype: :class:`int`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def xray_transition(self, xray_transition):  # pragma: no cover
        """
        Returns x-ray transition descriptor.

        {xray_transition}

        :return: x-ray transition descriptor
        :rtype: :class:`XrayTransition`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def xray_transition_notation(
        self, xray_transition, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        """
        Returns notation of an x-ray transition.

        {xray_transition}
        {notation}
        {encoding}
        {reference}

        :return: notation
        :rtype: :class:`str`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def xray_transition_energy_eV(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        """
        Returns energy of an element and X-ray transition (in eV).

        {element}
        {xray_transition}
        {reference}

        :return: energy (in eV)
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def xray_transition_probability(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        """
        Returns probability of an element and X-ray transition.

        {element}
        {xray_transition}
        {reference}

        :return: probability
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @formatdoc(**_docextras)
    def xray_transition_relative_weight(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        """
        Returns relative weight of an element and X-ray transition.

        {element}
        {xray_transition}
        {reference}

        :return: relative weight
        :rtype: :class:`float`
        {exception}
        """
        raise NotImplementedError

    @formatdoc(**_docextras)
    def xray_line(self, element, xray_transition):
        """
        Returns x-ray line descriptor.

        {element}
        {xray_transition}

        :return: x-ray line
        :rtype: :class:`XrayLine`
        {exception}
        """
        element = self.element(element)
        symbol = self.element_symbol(element)

        transition = self.xray_transition(xray_transition)

        iupac = "{} {}".format(
            symbol, self.xray_transition_notation(transition, "iupac", "utf16")
        )

        try:
            siegbahn = "{} {}".format(
                symbol, self.xray_transition_notation(transition, "siegbahn", "utf16")
            )
        except:
            siegbahn = iupac

        try:
            energy_eV = self.xray_transition_energy_eV(element, transition)
        except NotFound:
            energy_eV = None

        try:
            probability = self.xray_transition_probability(element, transition)
        except NotFound:
            probability = None

        try:
            relative_weight = self.xray_transition_relative_weight(element, transition)
        except NotFound:
            relative_weight = None

        return descriptor.XrayLine(
            element,
            transition,
            iupac,
            siegbahn,
            energy_eV,
            probability,
            relative_weight,
        )



""" """

# Standard library modules.
import os
import csv
import ast

# Third party modules.
import pytest
import sqlalchemy

# Local modules.
from pyxray.parser.notation import (
    AtomicSubshellNotationParser,
    KnownXrayTransitionNotationParser,
)
from pyxray.sql.build import SqlDatabaseBuilder
from pyxray.sql.data import SqlDatabase
import pyxray.descriptor as descriptor
from pyxray.base import NotFound

# Globals and constants variables.


class EpqDatabaseBuilder(SqlDatabaseBuilder):
    def _find_parsers(self):
        return [
            ("atomic subshell notation", AtomicSubshellNotationParser()),
            ("xray transition notation", KnownXrayTransitionNotationParser()),
        ]


@pytest.fixture
def database(tmp_path):
    engine = sqlalchemy.create_engine("sqlite:///" + str(tmp_path.joinpath("epq.sql")))

    builder = EpqDatabaseBuilder(engine)
    builder.build()

    return SqlDatabase(engine)


def test_epq_atomicsubshell_notation(database, testdatadir):
    filepath = os.path.join(testdatadir, "epq_atomicsubshell.csv")
    with open(filepath, "r") as fp:
        reader = csv.DictReader(fp)

        for row in reader:
            atomic_subshell = descriptor.AtomicSubshell(
                int(row["n"]), int(row["l"]), int(float(row["j"]) * 2)
            )
            assert (
                database.atomic_subshell_notation(atomic_subshell, "siegbahn")
                == row["notation"]
            )


def test_epq_xraytransition_notation(database, testdatadir):
    filepath = os.path.join(testdatadir, "epq_xraytransition.csv")
    with open(filepath, "r") as fp:
        reader = csv.DictReader(fp)

        for row in reader:
            source = database.atomic_subshell(row["source"])
            destination = database.atomic_subshell(row["destination"])
            transition = descriptor.XrayTransition(source, destination)

            expected = ast.literal_eval('"' + row["notation"] + '"')
            expected = expected.replace("p", "\u2032")

            try:
                actual = database.xray_transition_notation(transition, "siegbahn")
            except NotFound:
                continue

            actual = actual.strip("I")

            assert actual == expected



"""
Definition of properties.
"""

# Standard library modules.
import dataclasses

# Third party modules.

# Local modules.
from pyxray.descriptor import (
    Element,
    Reference,
    AtomicShell,
    AtomicSubshell,
    XrayTransition,
    Notation,
    Language,
)

# Globals and constants variables.


@dataclasses.dataclass(frozen=True)
class ElementSymbol:
    reference: Reference
    element: Element
    value: str

    def __post_init__(self):
        if len(self.value) == 0 or len(self.value) > 3:
            raise ValueError("Symbol should be between 1 and 3 characters")

        if not self.value[0].isupper():
            raise ValueError("Symbol should start with a capital letter")


@dataclasses.dataclass(frozen=True)
class ElementName:
    reference: Reference
    element: Element
    language: Language
    value: str

    def __post_init__(self):
        if not self.value:
            raise ValueError("A name must be specified")


@dataclasses.dataclass(frozen=True)
class ElementAtomicWeight:
    reference: Reference
    element: Element
    value: float

    def __post_init__(self):
        if self.value <= 0.0:
            raise ValueError("Value must be greater than 0.0")


@dataclasses.dataclass(frozen=True)
class ElementMassDensity:
    reference: Reference
    element: Element
    value_kg_per_m3: float

    def __post_init__(self):
        if self.value_kg_per_m3 <= 0.0:
            raise ValueError("Value must be greater than 0.0")


@dataclasses.dataclass(frozen=True)
class AtomicShellNotation:
    reference: Reference
    atomic_shell: AtomicShell
    notation: Notation
    ascii: str
    utf16: str
    html: str
    latex: str


@dataclasses.dataclass(frozen=True)
class AtomicSubshellNotation:
    reference: Reference
    atomic_subshell: AtomicSubshell
    notation: Notation
    ascii: str
    utf16: str
    html: str
    latex: str


@dataclasses.dataclass(frozen=True)
class AtomicSubshellBindingEnergy:
    reference: Reference
    element: Element
    atomic_subshell: AtomicSubshell
    value_eV: float


@dataclasses.dataclass(frozen=True)
class AtomicSubshellRadiativeWidth:
    reference: Reference
    element: Element
    atomic_subshell: AtomicSubshell
    value_eV: float


@dataclasses.dataclass(frozen=True)
class AtomicSubshellNonRadiativeWidth:
    reference: Reference
    element: Element
    atomic_subshell: AtomicSubshell
    value_eV: float


@dataclasses.dataclass(frozen=True)
class AtomicSubshellOccupancy:
    reference: Reference
    element: Element
    atomic_subshell: AtomicSubshell
    value: float


@dataclasses.dataclass(frozen=True)
class XrayTransitionNotation:
    reference: Reference
    xray_transition: XrayTransition
    notation: Notation
    ascii: str
    utf16: str
    html: str
    latex: str


@dataclasses.dataclass(frozen=True)
class XrayTransitionEnergy:
    reference: Reference
    element: Element
    xray_transition: XrayTransition
    value_eV: float


@dataclasses.dataclass(frozen=True)
class XrayTransitionProbability:
    reference: Reference
    element: Element
    xray_transition: XrayTransition
    value: float


@dataclasses.dataclass(frozen=True)
class XrayTransitionRelativeWeight:
    reference: Reference
    element: Element
    xray_transition: XrayTransition
    value: float



"""
Definition of descriptors.
"""

__all__ = [
    "Element",
    "AtomicShell",
    "AtomicSubshell",
    "XrayTransition",
    "XrayLine",
    "Language",
    "Notation",
    "Reference",
]

# Standard library modules.
import dataclasses
import typing
import functools
from collections.abc import Sequence

# Third party modules.

# Local modules.

# Globals and constants variables.


@dataclasses.dataclass(frozen=True, order=True)
class Element:
    atomic_number: int

    def __post_init__(self):
        if self.atomic_number < 1 or self.atomic_number > 118:
            raise ValueError(
                "Atomic number ({0}) must be [1, 118]".format(self.atomic_number)
            )

    def __repr__(self):
        return "{}(z={})".format(self.__class__.__name__, self.atomic_number)

    @property
    def z(self):
        return self.atomic_number


@dataclasses.dataclass(frozen=True, order=True)
class AtomicShell:
    principal_quantum_number: int

    def __post_init__(self):
        if self.principal_quantum_number < 1:
            raise ValueError(
                "Principal quantum number ({0}) must be [1, inf[".format(
                    self.principal_quantum_number
                )
            )

    def __repr__(self):
        return "{}(n={})".format(self.__class__.__name__, self.principal_quantum_number)

    @property
    def n(self):
        return self.principal_quantum_number


@dataclasses.dataclass(frozen=True, order=True)
class AtomicSubshell:
    principal_quantum_number: int
    azimuthal_quantum_number: int
    total_angular_momentum_nominator: int

    def __post_init__(self):
        if isinstance(self.principal_quantum_number, AtomicShell):
            object.__setattr__(
                self, "principal_quantum_number", self.principal_quantum_number.n
            )

        if (
            self.principal_quantum_number is not None
            and self.azimuthal_quantum_number is not None
        ):
            lmin = 0
            lmax = self.principal_quantum_number - 1
            if (
                self.azimuthal_quantum_number < lmin
                or self.azimuthal_quantum_number > lmax
            ):
                raise ValueError(
                    "Azimuthal quantum number ({0}) must be between [{1}, {2}]".format(
                        self.azimuthal_quantum_number, lmin, lmax
                    )
                )

        if (
            self.azimuthal_quantum_number is not None
            and self.total_angular_momentum_nominator is not None
        ):
            jmin_n = 2 * abs(self.azimuthal_quantum_number - 0.5)
            jmax_n = 2 * abs(self.azimuthal_quantum_number + 0.5)

            if (
                self.total_angular_momentum_nominator < jmin_n
                or self.total_angular_momentum_nominator > jmax_n
            ):
                raise ValueError(
                    "Total angular momentum ({0}) must be between [{1}, {2}]".format(
                        self.total_angular_momentum_nominator, jmin_n, jmax_n
                    )
                )

    def __repr__(self):
        def _format(value):
            return "*" if value is None else value

        return "{}(n={}, l={}, j={:.1f})".format(
            self.__class__.__name__, _format(self.n), _format(self.l), _format(self.j)
        )

    @property
    def atomic_shell(self):
        return AtomicShell(self.principal_quantum_number)

    @property
    def n(self):
        return self.principal_quantum_number

    @property
    def l(self):
        return self.azimuthal_quantum_number

    @property
    def j_n(self):
        return self.total_angular_momentum_nominator

    @property
    def total_angular_momentum(self):
        if self.total_angular_momentum_nominator is None:
            return None
        return self.total_angular_momentum_nominator / 2.0

    @property
    def j(self):
        return self.total_angular_momentum


@dataclasses.dataclass(frozen=True)
class XrayTransition:
    source_principal_quantum_number: int = None
    source_azimuthal_quantum_number: int = None
    source_total_angular_momentum_nominator: int = None
    destination_principal_quantum_number: int = None
    destination_azimuthal_quantum_number: int = None
    destination_total_angular_momentum_nominator: int = None

    def __init__(self, *args):
        if len(args) == 6:
            src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n = args

        elif len(args) == 2:
            if isinstance(args[0], Sequence):
                src_n, src_l, src_j_n = args[0]
            elif isinstance(args[0], AtomicSubshell):
                src_n = args[0].n
                src_l = args[0].l
                src_j_n = args[0].j_n
            else:
                raise ValueError("Unknown argument: {}".format(args[0]))

            if isinstance(args[1], Sequence):
                dst_n, dst_l, dst_j_n = args[1]
            elif isinstance(args[1], AtomicSubshell):
                dst_n = args[1].n
                dst_l = args[1].l
                dst_j_n = args[1].j_n
            else:
                raise ValueError("Unknown argument: {}".format(args[1]))

        elif len(args) == 4:
            if isinstance(args[0], Sequence):
                src_n, src_l, src_j_n = args[0]
                dst_n, dst_l, dst_j_n = args[1:]
            elif isinstance(args[0], AtomicSubshell):
                src_n = args[0].n
                src_l = args[0].l
                src_j_n = args[0].j_n
                dst_n, dst_l, dst_j_n = args[1:]
            elif isinstance(args[-1], Sequence):
                src_n, src_l, src_j_n = args[:-1]
                dst_n, dst_l, dst_j_n = args[-1]
            elif isinstance(args[-1], AtomicSubshell):
                src_n, src_l, src_j_n = args[:-1]
                dst_n = args[-1].n
                dst_l = args[-1].l
                dst_j_n = args[-1].j_n

        else:
            raise ValueError("Unsupported number of arguments: {}".format(len(args)))

        object.__setattr__(self, "source_principal_quantum_number", src_n)
        object.__setattr__(self, "source_azimuthal_quantum_number", src_l)
        object.__setattr__(self, "source_total_angular_momentum_nominator", src_j_n)
        object.__setattr__(self, "destination_principal_quantum_number", dst_n)
        object.__setattr__(self, "destination_azimuthal_quantum_number", dst_l)
        object.__setattr__(
            self, "destination_total_angular_momentum_nominator", dst_j_n
        )

    @classmethod
    def is_radiative(cls, source_subshell, destination_subshell):
        """
        Inspired from NIST EPQ library by Nicholas Ritchie.
        """

        def electric_dipole_permitted(n0, l0, j0_n, n1, l1, j1_n):
            delta_j_n = abs(j1_n - j0_n)
            if delta_j_n > 2:
                return False
            assert delta_j_n == 0 or delta_j_n == 2
            return abs(l1 - l0) == 1

        def electric_quadrupole_permitted(n0, l0, j0_n, n1, l1, j1_n):
            delta_j_n = abs(j1_n - j0_n)
            if delta_j_n > 4:
                return False
            if j0_n == 1 and j1_n == 1:
                return False
            assert delta_j_n == 0 or delta_j_n == 2 or delta_j_n == 4

            delta_l = abs(l1 - l0)
            return delta_l == 0 or delta_l == 2

        n0 = source_subshell.n
        l0 = source_subshell.l
        j0_n = source_subshell.j_n
        n1 = destination_subshell.n
        l1 = destination_subshell.l
        j1_n = destination_subshell.j_n

        if n0 == n1:
            return False

        if not (
            electric_dipole_permitted(n0, l0, j0_n, n1, l1, j1_n)
            or electric_quadrupole_permitted(n0, l0, j0_n, n1, l1, j1_n)
        ):
            return False

        return True

    def __repr__(self):
        def _format(value):
            return "*" if value is None else value

        return "{}([n={}, l={}, j={}] -> [n={}, l={}, j={}])".format(
            self.__class__.__name__,
            _format(self.source_principal_quantum_number),
            _format(self.source_azimuthal_quantum_number),
            _format(self.source_total_angular_momentum),
            _format(self.destination_principal_quantum_number),
            _format(self.destination_azimuthal_quantum_number),
            _format(self.destination_total_angular_momentum),
        )

    @property
    def source_subshell(self):
        return AtomicSubshell(
            self.source_principal_quantum_number,
            self.source_azimuthal_quantum_number,
            self.source_total_angular_momentum_nominator,
        )

    @property
    def destination_subshell(self):
        return AtomicSubshell(
            self.destination_principal_quantum_number,
            self.destination_azimuthal_quantum_number,
            self.destination_total_angular_momentum_nominator,
        )

    @property
    def source_total_angular_momentum(self):
        if self.source_total_angular_momentum_nominator is None:
            return None
        return self.source_total_angular_momentum_nominator / 2.0

    @property
    def destination_total_angular_momentum(self):
        if self.destination_total_angular_momentum_nominator is None:
            return None
        return self.destination_total_angular_momentum_nominator / 2.0


@dataclasses.dataclass(frozen=True)
class XrayLine:
    element: Element
    transition: XrayTransition
    iupac: str = dataclasses.field(compare=False)
    siegbahn: str = dataclasses.field(compare=False)
    energy_eV: float = dataclasses.field(compare=False)
    probability: float = dataclasses.field(default=None, compare=False)
    relative_weight: float = dataclasses.field(default=None, compare=False)

    def __post_init__(self):
        if not isinstance(self.element, Element):
            object.__setattr__(self, "element", Element(self.element))
        if not isinstance(self.transition, XrayTransition):
            object.__setattr__(self, "transition", XrayTransition(self.transition))

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.iupac)

    @property
    def atomic_number(self):
        return self.element.atomic_number

    @property
    def z(self):
        return self.element.atomic_number


@dataclasses.dataclass(frozen=True)
class Language:
    key: str

    def __post_init__(self):
        lencode = len(self.key)
        if lencode < 2 or lencode > 3:
            raise ValueError("Code must be between 2 and 3 characters")

        object.__setattr__(self, "key", self.key.lower())

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.key)


@dataclasses.dataclass(frozen=True)
class Notation:
    key: str

    def __post_init__(self):
        if not self.key:
            raise ValueError("Name cannot be empty")

        object.__setattr__(self, "key", self.key.lower())

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.key)


@dataclasses.dataclass(frozen=True)
class Reference:
    bibtexkey: str
    author: str = None
    year: str = None
    title: str = None
    type: str = None
    booktitle: str = None
    editor: str = None
    pages: str = None
    edition: str = None
    journal: str = None
    school: str = None
    address: str = None
    url: str = None
    note: str = None
    number: str = None
    series: str = None
    volume: str = None
    publisher: str = None
    organization: str = None
    chapter: str = None
    howpublished: str = None
    doi: str = None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.bibtexkey)



"""
Parsers for DTSA data.
"""

# Standard library modules.
import os
import csv
import logging
import pkgutil

# Third party modules.

# Local modules.
from pyxray.descriptor import Reference, Element, XrayTransition
from pyxray.property import (
    AtomicSubshellBindingEnergy,
    XrayTransitionEnergy,
    XrayTransitionRelativeWeight,
)
import pyxray.parser.base as base

# Globals and constants variables.
logger = logging.getLogger(__name__)

# noinspection PyArgumentList
DTSA1992 = Reference(
    "dtsa1992",
    author="C.E. Fiori and C.R. Swyt and R.L. Myklebust",
    year=1992,
    title="NIST/NIH Desk Top Spectrum Analyzer",
    url="https://www.cstl.nist.gov/div837/Division/outputs/DTSA/oldDTSA.htm",
)

# noinspection SpellCheckingInspection
_SUBSHELL_LOOKUP = {
    "Kedge": base.K,
    "LIedge": base.L1,
    "LIIedge": base.L2,
    "LIIIedge": base.L3,
    "MIedge": base.M1,
    "MIIedge": base.M2,
    "MIIIedge": base.M3,
    "MIVedge": base.M4,
    "MVedge": base.M5,
    "NIedge": base.N1,
    "NIIedge": base.N2,
    "NIIIedge": base.N3,
    "NIVedge": base.N4,
    "NVedge": base.N5,
    "NVIedge": base.N6,
    "NVIIedge": base.N7,
    "OIedge": base.O1,
}

# noinspection PyArgumentList
class DtsaSubshellParser(base._Parser):
    def __iter__(self):
        relative_path = os.path.join("..", "data", "dtsa_subshell.csv")
        content = pkgutil.get_data(__name__, relative_path).decode("utf8")
        reader = csv.reader(content.splitlines())

        # skip first line
        next(reader)

        subshell_data = []
        for row in reader:
            atomic_number = int(row[0])
            # noinspection PyPep8Naming
            energy_eV = float(row[1])
            subshell = str(row[2])

            subshell_data.append([atomic_number, energy_eV, subshell])

        length = len(subshell_data)
        for atomic_number, energy_eV, subshell_dtsa in subshell_data:

            subshell = _SUBSHELL_LOOKUP[subshell_dtsa]
            element = Element(atomic_number)
            prop = AtomicSubshellBindingEnergy(DTSA1992, element, subshell, energy_eV)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((atomic_number - 1) / length * 100.0))
            yield prop


_TRANSITION_LOOKUP = {
    "Ka1": base.Ka1,
    "Ka2": base.Ka2,
    "Kb1": base.Kb1,
    "Kb2": base.Kb2,
    "Kb3": base.Kb3,
    "Kb4": base.Kb4,
    "Kb5": base.Kb5,
    "La1": base.La1,
    "La2": base.La2,
    "Lb1": base.Lb1,
    "Lb2": base.Lb2,
    "Lb3": base.Lb3,
    "Lb4": base.Lb4,
    "Lb5": base.Lb5,
    "Lb6": base.Lb6,
    "Lb7": base.Lb7,
    "Lb9": base.Lb9,
    "Lb10": base.Lb10,
    "Lb15": base.Lb15,
    "Lb17": base.Lb17,
    "Leta": base.Leta,
    "Lg1": base.Lg1,
    "Lg2": base.Lg2,
    "Lg3": base.Lg3,
    "Lg4": base.Lg4,
    "Lg5": base.Lg5,
    "Lg6": base.Lg6,
    "Lg8": base.Lg8,
    "Lg11": base.Lg11,
    "Ll": base.Ll,
    "Ls": base.Ls,
    "Lt": base.Lt,
    "Lv": base.Lv,
    "Lu": base.Lu,
    "Ma1": base.Ma1,
    "Ma2": base.Ma2,
    "Mb": base.Mb,
    "Mg": base.Mg,
    "Mz1": base.Mz1,
    "Mz2": base.Mz2,
    "N4-N6": XrayTransition(base.N6, base.N4),
    "N5-N6": XrayTransition(base.N6, base.N5),
    #'Kb2': 'Kb2I',
    #'Kb4': 'Kb4I',
    #'Kb5': 'Kb5I',
    # SKa,SKa',SKa'',SKa3,SKa3',SKa3'',SKa4,SKa5,SKa6,SKa7,SKa8,SKa9,SKb',SKb'',SKb+4,SKb+5,SKb7,SKb8,SKb9,SKbN,SKbX,SLa',SLa+IX,SLa+X,SLa+Y,SLa1+Z,SLa2',SLa3,SLa3+Z,SLa4,SLa5,SLa6,SLa7,SLa8,SLa9,SLaa,SLas,SLb'',SLb1+4,SLb14,SLb2+1,SLb2+2,SLb2+3,SLb2+4,SLb2+5,SLb2+7,SLb2+A,SLb2+B,SLb2+C,SLb5+1,SLb5+2,SLg1',SLg10,SLg2',SLg2'',SLg9,SMa+1,SMa+2,SMa+3,SMa+4,SMb1,SMb2,SMb3,SMg',Skb,Skb10,Skb6
    "L1M1": XrayTransition(base.L1, base.M1),
    "L1N1": XrayTransition(base.N1, base.L1),
    "L1N4": XrayTransition(base.N4, base.L1),
    "L1O1": XrayTransition(base.O1, base.L1),
    "L1O4": XrayTransition(base.O4, base.L1),
    "L2M2": XrayTransition(base.L2, base.M2),
    "L2M5": XrayTransition(base.M5, base.L2),
    "L2N2": XrayTransition(base.N2, base.L2),
    "L2N3": XrayTransition(base.N3, base.L2),
    "L2N5": XrayTransition(base.N5, base.L2),
    "L2O2": XrayTransition(base.O2, base.L2),
    "L2O3": XrayTransition(base.O3, base.L2),
    "L2P2": XrayTransition(base.P2, base.L2),
    "L2P3": XrayTransition(base.P3, base.L2),
    "L3N2": XrayTransition(base.N2, base.L3),
    "L3N3": XrayTransition(base.N3, base.L3),
    "L3O2": XrayTransition(base.O2, base.L3),
    "L3O3": XrayTransition(base.O3, base.L3),
    "L3P1": XrayTransition(base.P1, base.L3),
    "M2O4": XrayTransition(base.O4, base.M2),
    "M1N2": XrayTransition(base.N2, base.M1),
    "M1N3": XrayTransition(base.N3, base.M1),
    "M2M4": XrayTransition(base.M4, base.M2),
    "M2N1": XrayTransition(base.N1, base.M2),
    "M2N4": XrayTransition(base.N4, base.M2),
    "M3M4": XrayTransition(base.M4, base.M3),
    "M3M5": XrayTransition(base.M5, base.M3),
    "M3N1": XrayTransition(base.N1, base.M3),
    "M3N4": XrayTransition(base.N4, base.M3),
    "M3O1": XrayTransition(base.O1, base.M3),
    "M3O4": XrayTransition(base.O4, base.M3),
    "M3O5": XrayTransition(base.O5, base.M3),
    "M4N3": XrayTransition(base.N3, base.M4),
    "M4O2": XrayTransition(base.O2, base.M4),
    "M5O3": XrayTransition(base.O3, base.M5),
}

# noinspection PyArgumentList
class DtsaLineParser(base._Parser):
    def __iter__(self):
        relative_path = os.path.join("..", "data", "dtsa_line.csv")
        content = pkgutil.get_data(__name__, relative_path).decode("utf8")
        reader = csv.reader(content.splitlines())

        # skip first line
        next(reader)

        line_data = []
        for row in reader:
            try:
                atomic_number = int(row[0])
                # noinspection PyPep8Naming
                energy_eV = float(row[1])
                fraction = float(row[2])
                line_label = str(row[3])

                line_data.append([atomic_number, energy_eV, fraction, line_label])
            except ValueError:
                pass

        unparse_lines = set()
        length = 2 * len(line_data)
        for atomic_number, energy_eV, fraction, line_label in line_data:
            try:
                transition = _TRANSITION_LOOKUP[line_label]
            except KeyError:
                #                logger.debug('Line not found: {} for {}'.format(line_label, atomic_number))
                unparse_lines.add(line_label)
                continue

            element = Element(atomic_number)

            prop = XrayTransitionEnergy(DTSA1992, element, transition, energy_eV)
            #                logger.debug('Parsed: {0}'.format(prop))
            self.update(int((atomic_number - 1) / length * 100.0))
            yield prop

            prop = XrayTransitionRelativeWeight(DTSA1992, element, transition, fraction)
            #                logger.debug('Parsed: {0}'.format(prop))
            self.update(int((atomic_number - 1) / length * 100.0))
            yield prop

        logger.debug("Lines not found: {}".format(sorted(unparse_lines)))



#!/usr/bin/env python
"""
Parser for the Evaluated Atomic Data Library (EADL)
"""

# Standard library modules.
import os
import re
import logging

# Third party modules.
import requests

# Local modules.
from pyxray.descriptor import Reference, Element, AtomicSubshell, XrayTransition
from pyxray.property import (
    ElementAtomicWeight,
    AtomicSubshellBindingEnergy,
    AtomicSubshellRadiativeWidth,
    AtomicSubshellNonRadiativeWidth,
    AtomicSubshellOccupancy,
    XrayTransitionEnergy,
    XrayTransitionProbability,
)
import pyxray.parser.base as base

# Globals and constants variables.
logger = logging.getLogger(__name__)

PERKINS1991 = Reference(
    "perkins1991",
    author="Perkins, S.T. and Cullen, D.E.",
    title="Tables and Graphs of Atomic Subshell and Relaxation Data Derived from the LLNL Evaluated Atomic Data Library (EADL), Z = 1 - 100",
    organization="Lawrence Livermore National Laboratory",
    year=1991,
    volume=30,
)

EADL_URL = "https://www-nds.iaea.org/epdl97/data/endl/eadl/eadl.all"

FLOAT_PATTERN = re.compile(r"([\d.]+)([\+\-])([\d ]+)")


def float_(text):
    text = text.strip()
    base, sign, exp = FLOAT_PATTERN.match(text).groups()
    base = float(base)
    exp = float(exp) if sign == "+" else -float(exp)
    return base * 10 ** exp


# Conversion from EADL to AtomicSubshell descriptor
ATOMIC_SUBSHELLS = {
    # K
    1: base.K,
    # L
    3: base.L1,
    5: base.L2,
    6: base.L3,
    # M
    8: base.M1,
    10: base.M2,
    11: base.M3,
    13: base.M4,
    14: base.M5,
    # N
    16: base.N1,
    18: base.N2,
    19: base.N3,
    21: base.N4,
    22: base.N5,
    24: base.N6,
    25: base.N7,
    # O
    27: base.O1,
    29: base.O2,
    30: base.O3,
    32: base.O4,
    33: base.O5,
    35: base.O6,
    36: base.O7,
    38: base.O8,
    39: base.O9,
    # P
    41: base.P1,
    43: base.P2,
    44: base.P3,
    46: base.P4,
    47: base.P5,
    49: base.P6,
    50: base.P7,
    52: base.P8,
    53: base.P9,
    55: base.P10,
    56: base.P11,
    # Q
    58: base.Q1,
    60: base.Q2,
    61: base.Q3,
}

# C
REACTION_DESCRIPTOR_SUBSHELL = 91
REACTION_DESCRIPTOR_TRANSITION = 92

# I
REACTION_PROPERTY_NUMBER_ELECTRON = 912
REACTION_PROPERTY_BINDING_ENERGY = 913
REACTION_PROPERTY_RADIATIVE_WIDTH = 921
REACTION_PROPERTY_NONRADIATIVE_WIDTH = 922
REACTION_PROPERTY_RADIATIVE_PROBABILITY = 931
REACTION_PROPERTY_NONRADIATIVE_PROBABILITY = 932

# S
REACTION_MODIFIER_NONE = 0
REACTION_MODIFIER_EXTRA = 91

# Yo
OUTGOING_PARTICLE_NONE = 0
OUTGOING_PARTICLE_PHOTON = 7
OUTGOING_PARTICLE_ELECTRON = 9

SEPERATOR = "                                                                       1"

MAX_Z = 100


class Perkins1991Parser(base._Parser):
    def __iter__(self):
        r = requests.get(EADL_URL, stream=True, verify=False)

        try:
            rows = []
            for line in r.iter_lines():
                line = line.decode("ascii")

                if line == SEPERATOR:
                    yield from self._iter_rows(rows)

                    element = self._extract_element(rows)
                    self.update(int(element.z / MAX_Z * 100.0))

                    rows = []
                    continue

                rows.append(line)

        finally:
            r.close()

    def _iter_rows(self, rows):
        yield from self._parse_atomic_weight(rows)
        yield from self._parse_atomic_subshell_binding_energy(rows)
        yield from self._parse_atomic_subshell_radiative_width(rows)
        #        yield from self._parse_atomic_subshell_nonradiative_width(rows)
        yield from self._parse_atomic_subshell_occupaqncy(rows)
        yield from self._parse_radiative_transition(rows)

    #        yield from self._parse_nonradiative_transition(rows)

    def _extract_element(self, rows):
        return Element(int(rows[0][0:3]))

    def _extract_reaction_descriptor(self, rows):
        return int(rows[1][0:2])

    def _extract_reaction_modifier(self, rows):
        return int(rows[1][5:8])

    def _extract_reaction_property(self, rows):
        return int(rows[1][2:5])

    def _extract_outgoing_particle(self, rows):
        return int(rows[0][10:12])

    def _extract_source_subshell(self, rows):
        return ATOMIC_SUBSHELLS[int(float_(rows[0][0:11]))]

    def _extract_secondary_destination_subshell(self, rows):
        return ATOMIC_SUBSHELLS[int(float_(rows[0][11:22]))]

    def _extract_destination_subshell(self, rows):
        return ATOMIC_SUBSHELLS[int(float_(rows[1][21:32]))]

    def _parse_atomic_weight(self, rows):
        # Note: Fake descriptor and property. Selected just in order to only
        # have one entry for the atomic weight
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_SUBSHELL:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_NUMBER_ELECTRON:
            return

        element = self._extract_element(rows)
        value = float_(rows[0][13:24])
        prop = ElementAtomicWeight(PERKINS1991, element, value)
        logger.debug("Parsed: {0}".format(prop))
        yield prop

    def _parse_atomic_subshell_binding_energy(self, rows):
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_SUBSHELL:
            return

        reaction_modifier = self._extract_reaction_modifier(rows)
        if reaction_modifier != REACTION_MODIFIER_NONE:
            return

        outgoing_particle = self._extract_outgoing_particle(rows)
        if outgoing_particle != OUTGOING_PARTICLE_NONE:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_BINDING_ENERGY:
            return

        element = self._extract_element(rows)

        for row in rows[2:]:
            atomic_subshell = self._extract_source_subshell([row])
            value_eV = float_(row[11:22]) * 1e6
            prop = AtomicSubshellBindingEnergy(
                PERKINS1991, element, atomic_subshell, value_eV
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

    def _parse_atomic_subshell_radiative_width(self, rows):
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_SUBSHELL:
            return

        reaction_modifier = self._extract_reaction_modifier(rows)
        if reaction_modifier != REACTION_MODIFIER_NONE:
            return

        outgoing_particle = self._extract_outgoing_particle(rows)
        if outgoing_particle != OUTGOING_PARTICLE_NONE:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_RADIATIVE_WIDTH:
            return

        element = self._extract_element(rows)

        for row in rows[2:]:
            atomic_subshell = self._extract_source_subshell([row])
            value_eV = float_(row[11:22]) * 1e6
            prop = AtomicSubshellRadiativeWidth(
                PERKINS1991, element, atomic_subshell, value_eV
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

    def _parse_atomic_subshell_nonradiative_width(self, rows):
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_SUBSHELL:
            return

        reaction_modifier = self._extract_reaction_modifier(rows)
        if reaction_modifier != REACTION_MODIFIER_NONE:
            return

        outgoing_particle = self._extract_outgoing_particle(rows)
        if outgoing_particle != OUTGOING_PARTICLE_NONE:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_NONRADIATIVE_WIDTH:
            return

        element = self._extract_element(rows)

        for row in rows[2:]:
            atomic_subshell = self._extract_source_subshell([row])
            value_eV = float_(row[11:22]) * 1e6
            prop = AtomicSubshellNonRadiativeWidth(
                PERKINS1991, element, atomic_subshell, value_eV
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

    def _parse_atomic_subshell_occupaqncy(self, rows):
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_SUBSHELL:
            return

        reaction_modifier = self._extract_reaction_modifier(rows)
        if reaction_modifier != REACTION_MODIFIER_NONE:
            return

        outgoing_particle = self._extract_outgoing_particle(rows)
        if outgoing_particle != OUTGOING_PARTICLE_NONE:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_NUMBER_ELECTRON:
            return

        element = self._extract_element(rows)

        for row in rows[2:]:
            atomic_subshell = self._extract_source_subshell([row])
            value = int(float_(row[11:22]))
            prop = AtomicSubshellOccupancy(PERKINS1991, element, atomic_subshell, value)
            logger.debug("Parsed: {0}".format(prop))
            yield prop

    def _parse_radiative_transition(self, rows):
        reaction_descriptor = self._extract_reaction_descriptor(rows)
        if reaction_descriptor != REACTION_DESCRIPTOR_TRANSITION:
            return

        reaction_modifier = self._extract_reaction_modifier(rows)
        if reaction_modifier != REACTION_MODIFIER_EXTRA:
            return

        outgoing_particle = self._extract_outgoing_particle(rows)
        if outgoing_particle != OUTGOING_PARTICLE_PHOTON:
            return

        reaction_property = self._extract_reaction_property(rows)
        if reaction_property != REACTION_PROPERTY_RADIATIVE_PROBABILITY:
            return

        element = self._extract_element(rows)

        destination_subshell = self._extract_destination_subshell(rows)

        for row in rows[2:]:
            source_subshell = self._extract_source_subshell([row])
            transition = XrayTransition(source_subshell, destination_subshell)

            value = float_(row[11:22])
            prop = XrayTransitionProbability(PERKINS1991, element, transition, value)
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            value_eV = float_(row[22:33]) * 1e6
            prop = XrayTransitionEnergy(PERKINS1991, element, transition, value_eV)
            logger.debug("Parsed: {0}".format(prop))
            yield prop


#    def _parse_nonradiative_transition(self, rows):
#        reaction_descriptor = self._extract_reaction_descriptor(rows)
#        if reaction_descriptor != REACTION_DESCRIPTOR_TRANSITION:
#            return
#
#        reaction_modifier = self._extract_reaction_modifier(rows)
#        if reaction_modifier != REACTION_MODIFIER_EXTRA:
#            return
#
#        outgoing_particle = self._extract_outgoing_particle(rows)
#        if outgoing_particle != OUTGOING_PARTICLE_ELECTRON:
#            return
#
#        reaction_property = self._extract_reaction_property(rows)
#        if reaction_property != REACTION_PROPERTY_NONRADIATIVE_PROBABILITY:
#            return
#
#        element = self._extract_element(rows)
#
#        destination_subshell = self._extract_destination_subshell(rows)
#
#        for row in rows[2:]:
#            source_subshell = self._extract_source_subshell([row])
#            secondary_destination_subshell = \
#                self._extract_secondary_destination_subshell([row])
#            transition = Transition(source_subshell, destination_subshell,
#                                    secondary_destination_subshell)
#
#            value = float_(row[22:33])
#            prop = TransitionProbability(PERKINS1991, element, transition, value)
#            logger.debug('Parsed: {0}'.format(prop))
#            yield prop
#
#            value_eV = float_(row[33:44]) * 1e6
#            prop = TransitionEnergy(PERKINS1991, element, transition, value_eV)
#            logger.debug('Parsed: {0}'.format(prop))
#            yield prop



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.

# Local modules.
from pyxray.parser.jeol import JEOLTransitionParser

# Globals and constants variables.


def test_jeol():
    parser = JEOLTransitionParser()
    assert len(list(parser)) == 2372



"""

"""

# Standard library modules.
import logging

# Third party modules.

# Local modules.
import pyxray.parser.base as base
from pyxray.descriptor import (
    Element,
    Reference,
    XrayTransition,
    AtomicShell,
    AtomicSubshell,
    Notation,
)
from pyxray.property import (
    ElementSymbol,
    AtomicShellNotation,
    AtomicSubshellNotation,
    XrayTransitionNotation,
)

# Globals and constants variables.
logger = logging.getLogger(__name__)

JENKINS1991 = Reference(
    "jenkins1991",
    author="Jenkins, R. and Manne, R. and Robin, R. and Senemaud, C.",
    journal="X-Ray Spectrometry",
    doi="10.1002/xrs.1300200308",
    pages="149--155",
    title="{IUPAC} --- nomenclature system for x-ray spectroscopy",
    volume=20,
    year=1991,
)

BEARDEN1967 = Reference(
    "bearden1967",
    author="Bearden, J. A.",
    journal="Rev. Mod. Phys.",
    doi="10.1103/RevModPhys.39.78",
    pages="78--124",
    title="X-Ray Wavelengths",
    volume=39,
    year=1967,
)


class ElementSymbolParser(base._Parser):

    SYMBOLS = [
        "H",
        "He",
        "Li",
        "Be",
        "B",
        "C",
        "N",
        "O",
        "F",
        "Ne",
        "Na",
        "Mg",
        "Al",
        "Si",
        "P",
        "S",
        "Cl",
        "Ar",
        "K",
        "Ca",
        "Sc",
        "Ti",
        "V",
        "Cr",
        "Mn",
        "Fe",
        "Co",
        "Ni",
        "Cu",
        "Zn",
        "Ga",
        "Ge",
        "As",
        "Se",
        "Br",
        "Kr",
        "Rb",
        "Sr",
        "Y",
        "Zr",
        "Nb",
        "Mo",
        "Tc",
        "Ru",
        "Rh",
        "Pd",
        "Ag",
        "Cd",
        "In",
        "Sn",
        "Sb",
        "Te",
        "I",
        "Xe",
        "Cs",
        "Ba",
        "La",
        "Ce",
        "Pr",
        "Nd",
        "Pm",
        "Sm",
        "Eu",
        "Gd",
        "Tb",
        "Dy",
        "Ho",
        "Er",
        "Tm",
        "Yb",
        "Lu",
        "Hf",
        "Ta",
        "W",
        "Re",
        "Os",
        "Ir",
        "Pt",
        "Au",
        "Hg",
        "Tl",
        "Pb",
        "Bi",
        "Po",
        "At",
        "Rn",
        "Fr",
        "Ra",
        "Ac",
        "Th",
        "Pa",
        "U",
        "Np",
        "Pu",
        "Am",
        "Cm",
        "Bk",
        "Cf",
        "Es",
        "Fm",
        "Md",
        "No",
        "Lr",
        "Rf",
        "Db",
        "Sg",
        "Bh",
        "Hs",
        "Mt",
        "Ds",
        "Rg",
        "Cn",
        "Uut",
        "Fl",
        "Uup",
        "Lv",
        "Uus",
        "Uuo",
    ]

    def __iter__(self):
        length = len(self.SYMBOLS)
        for z, symbol in enumerate(self.SYMBOLS, 1):
            prop = ElementSymbol(base.UNATTRIBUTED, Element(z), symbol)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((z - 1) / length * 100.0))
            yield prop


MAX_N = 7


def iter_subshells(max_n):
    nprev = 0
    for n in range(1, max_n + 1):
        if nprev != n:
            i = 1

        for l in range(0, n):
            for j in set([abs(l + 0.5), abs(l - 0.5)]):
                j_n = int(j * 2)
                yield n, l, j_n, i
                i += 1
                nprev = n


def iter_transitions(max_n):
    for src_n, src_l, src_j_n, src_i in iter_subshells(max_n):

        for dst_n, dst_l, dst_j_n, dst_i in iter_subshells(MAX_N):
            # Cannot transition to itself
            if src_n == dst_n and src_l == dst_l and src_j_n == dst_j_n:
                continue

            # Transition must be from more to less energytic shell or
            # within the same shell
            if src_n < dst_n:
                continue

            # Coster-Kroning transition must be from more to less energetic subshells
            if src_n == dst_n and src_i <= dst_i:
                continue

            yield XrayTransition(
                src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n
            ), src_i, dst_i


class AtomicShellNotationParser(base._Parser):

    SIEGBAHN_SHELLS = ["K", "L", "M", "N", "O", "P", "Q"]
    IUPAC_SHELLS = ["K", "L", "M", "N", "O", "P", "Q"]
    ORBITAL_SHELLS = ["1", "2", "3", "4", "5", "6", "7"]

    @classmethod
    def _create_entry_siegbahn(cls, n):
        s = cls.SIEGBAHN_SHELLS[n - 1]
        return s, s, s, s

    @classmethod
    def _create_entry_iupac(cls, n):
        s = cls.IUPAC_SHELLS[n - 1]
        return s, s, s, s

    @classmethod
    def _create_entry_orbital(cls, n):
        s = cls.ORBITAL_SHELLS[n - 1]
        return s, s, s, s

    def __iter__(self):
        length = MAX_N * 3
        progress = 0
        for n in range(1, MAX_N + 1):
            self.update(int(progress / length * 100.0))
            progress += 1

            atomic_shell = AtomicShell(n)

            ascii, utf16, html, latex = self._create_entry_siegbahn(n)
            prop = AtomicShellNotation(
                base.UNATTRIBUTED,
                atomic_shell,
                base.SIEGBAHN,
                ascii,
                utf16,
                html,
                latex,
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            ascii, utf16, html, latex = self._create_entry_iupac(n)
            prop = AtomicShellNotation(
                base.UNATTRIBUTED, atomic_shell, base.IUPAC, ascii, utf16, html, latex
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            ascii, utf16, html, latex = self._create_entry_orbital(n)
            prop = AtomicShellNotation(
                base.UNATTRIBUTED, atomic_shell, base.ORBITAL, ascii, utf16, html, latex
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop


class AtomicSubshellNotationParser(base._Parser):

    SIEGBAHN_NUMBERS = [
        "I",
        "II",
        "III",
        "IV",
        "V",
        "VI",
        "VII",
        "VIII",
        "IX",
        "X",
        "XI",
        "XII",
        "XIII",
    ]
    ORBITAL_L = ["s", "p", "d", "f", "g", "h", "i"]

    @classmethod
    def _create_entry_siegbahn(cls, n, l, j_n, i):
        shell = AtomicShellNotationParser.SIEGBAHN_SHELLS[n - 1]
        s = shell
        if n != 1:  # No KI
            s += cls.SIEGBAHN_NUMBERS[i - 1]
        return s, s, s, s

    @classmethod
    def _create_entry_iupac(cls, n, l, j_n, i):
        shell = AtomicShellNotationParser.IUPAC_SHELLS[n - 1]
        s = html = latex = shell
        if n != 1:  # No K1
            s += str(i)
            html += "<sub>{0:d}</sub>".format(i)
            latex += "$_{{{0:d}}}$".format(i)
        return s, s, html, latex

    @classmethod
    def _create_entry_orbital(cls, n, l, j_n, i):
        l_letter = cls.ORBITAL_L[l]
        s = "{0:d}{1}{2}/2".format(n, l_letter, j_n)
        html = "{0:d}{1}<sub>{2}/2</sub>".format(n, l_letter, j_n)
        latex = "{0:d}{1}$_{{{2}/2}}$".format(n, l_letter, j_n)
        return s, s, html, latex

    def __iter__(self):
        for n, l, j_n, i in iter_subshells(MAX_N):
            self.update(int(n / MAX_N * 100.0))

            atomic_subshell = AtomicSubshell(n, l, j_n)

            ascii, utf16, html, latex = self._create_entry_siegbahn(n, l, j_n, i)
            prop = AtomicSubshellNotation(
                base.UNATTRIBUTED,
                atomic_subshell,
                base.SIEGBAHN,
                ascii,
                utf16,
                html,
                latex,
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            ascii, utf16, html, latex = self._create_entry_iupac(n, l, j_n, i)
            prop = AtomicSubshellNotation(
                base.UNATTRIBUTED,
                atomic_subshell,
                base.IUPAC,
                ascii,
                utf16,
                html,
                latex,
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            ascii, utf16, html, latex = self._create_entry_orbital(n, l, j_n, i)
            prop = AtomicSubshellNotation(
                base.UNATTRIBUTED,
                atomic_subshell,
                base.ORBITAL,
                ascii,
                utf16,
                html,
                latex,
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop


class GenericXrayTransitionNotationParser(base._Parser):
    def __iter__(self):
        length = (MAX_N * MAX_N) ** 2
        progress = 0
        for transition, src_i, dst_i in iter_transitions(MAX_N):
            progress += 1
            self.update(int(progress / length * 100.0))

            src = transition.source_subshell
            dst = transition.destination_subshell

            (
                ascii0,
                utf0,
                html0,
                latex0,
            ) = AtomicSubshellNotationParser._create_entry_iupac(
                src.n, src.l, src.j_n, src_i
            )
            (
                ascii1,
                utf1,
                html1,
                latex1,
            ) = AtomicSubshellNotationParser._create_entry_iupac(
                dst.n, dst.l, dst.j_n, dst_i
            )

            ascii = "{0}-{1}".format(ascii1, ascii0)
            utf = "{0}\u2013{1}".format(utf1, utf0)
            html = "{0}&ndash;{1}".format(html1, html0)
            latex = "{0}--{1}".format(latex1, latex0)

            prop = XrayTransitionNotation(
                base.UNATTRIBUTED, transition, base.IUPAC, ascii, utf, html, latex
            )
            logger.debug("Parsed: {0}".format(prop))
            yield prop


class KnownXrayTransitionNotationParser(base._Parser):
    def __iter__(self):
        assert base.Ka1.source_subshell == base.L3
        assert base.Ka1.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ka1,
            base.SIEGBAHN,
            "Ka1",
            "K\u03b11",
            "K&alpha;<sub>1</sub>",
            "\\ensuremath{\\mathrm{K}\\alpha_1}",
        )

        assert base.Ka2.source_subshell == base.L2
        assert base.Ka2.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ka2,
            base.SIEGBAHN,
            "Ka2",
            "K\u03b12",
            "K&alpha;<sub>2</sub>",
            "\\ensuremath{\\mathrm{K}\\alpha_2}",
        )

        assert base.Ka.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ka,
            base.SIEGBAHN,
            "Ka",
            "K\u03b1",
            "K&alpha;",
            "\\ensuremath{\\mathrm{K}\\alpha}",
        )
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ka,
            base.IUPAC,
            "K-L(2,3)",
            "K\u2013L(2,3)",
            "K&ndash;L<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{K}}--\\ensuremath{\\mathrm{L}_{2,3}}",
        )

        assert base.Kb1.source_subshell == base.M3
        assert base.Kb1.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb1,
            base.SIEGBAHN,
            "Kb1",
            "K\u03b21",
            "K&beta;<sub>1</sub>",
            "\\ensuremath{\\mathrm{K}\\beta_1}",
        )

        assert base.Kb2_1.source_subshell == base.N3
        assert base.Kb2_1.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb2_1,
            base.SIEGBAHN,
            "Kb2I",
            "K\u03b22I",
            "K&beta;<sub>2</sub><sup>I</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_2^I}",
        )

        assert base.Kb2_2.source_subshell == base.N2
        assert base.Kb2_2.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb2_2,
            base.SIEGBAHN,
            "Kb2II",
            "K\u03b22II",
            "K&beta;<sub>2</sub><sup>II</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_2^{II}}",
        )

        assert base.Kb2.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb2,
            base.SIEGBAHN,
            "Kb2",
            "K\u03b22",
            "K&beta;<sub>2</sub>",
            "\\ensuremath{\\mathrm{K}\\beta_2}",
        )

        assert base.Kb3.source_subshell == base.M2
        assert base.Kb3.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb3,
            base.SIEGBAHN,
            "Kb3",
            "K\u03b23",
            "K&beta;<sub>3</sub>",
            "\\ensuremath{\\mathrm{K}\\beta_3}",
        )

        assert base.Kb1_3.destination_subshell == base.K
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Kb1_3,
            base.SIEGBAHN,
            "Kb1,3",
            "K\u03b21,3",
            "K&beta;<sup>1,3</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_{1,3}}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Kb1_3,
            base.IUPAC,
            "K-M(2,3)",
            "K\u2013M(2,3)",
            "K&ndash;M<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{K}}--\\ensuremath{\\mathrm{M}_{2,3}}",
        )

        assert base.Kb4_1.source_subshell == base.N5
        assert base.Kb4_1.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb4_1,
            base.SIEGBAHN,
            "Kb4I",
            "K\u03b24II",
            "K&beta;<sub>4</sub><sup>I</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_4^I}",
        )

        assert base.Kb4_2.source_subshell == base.N4
        assert base.Kb4_2.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb4_2,
            base.SIEGBAHN,
            "Kb4II",
            "K\u03b24II",
            "K&beta;<sub>4</sub><sup>II</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_4^{II}}",
        )

        assert base.Kb4.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb4,
            base.SIEGBAHN,
            "Kb4",
            "K\u03b24",
            "K&beta;<sub>4</sub>",
            "\\ensuremath{\\mathrm{K}\\beta_4}",
        )

        assert base.Kb5_1.source_subshell == base.M5
        assert base.Kb5_1.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb5_1,
            base.SIEGBAHN,
            "Kb5I",
            "K\u03b25I",
            "K&beta;<sub>5</sub><sup>I</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_5^I}",
        )

        assert base.Kb5_2.source_subshell == base.M4
        assert base.Kb5_2.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb5_2,
            base.SIEGBAHN,
            "Kb5II",
            "K\u03b25II",
            "K&beta;<sub>5</sub><sup>II</sup>",
            "\\ensuremath{\\mathrm{K}\\beta_5^{II}}",
        )

        assert base.Kb5.destination_subshell == base.K
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Kb5,
            base.SIEGBAHN,
            "Kb5",
            "K\u03b25",
            "K&beta;<sub>5</sub>",
            "\\ensuremath{\\mathrm{K}\\beta_5}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Kb5,
            base.IUPAC,
            "K-M(4,5)",
            "K\u2013M(4,5)",
            "K&ndash;M<sub>4,5</sub>",
            "\\ensuremath{\\mathrm{K}}--\\ensuremath{\\mathrm{M}_{4,5}}",
        )

        assert base.KO2_3.destination_subshell == base.K
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.KO2_3,
            base.IUPAC,
            "K-O(2,3)",
            "K\u2013O(2,3)",
            "K&ndash;O<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{K}}--\\ensuremath{\\mathrm{O}_{2,3}}",
        )

        assert base.La1.source_subshell == base.M5
        assert base.La1.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.La1,
            base.SIEGBAHN,
            "La1",
            "L\u03b11",
            "L&alpha;<sub>1</sub>",
            "\\ensuremath{\\mathrm{L}\\alpha_1}",
        )

        assert base.La2.source_subshell == base.M4
        assert base.La2.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.La2,
            base.SIEGBAHN,
            "La2",
            "L\u03b12",
            "L&alpha;<sub>2</sub>",
            "\\ensuremath{\\mathrm{L}\\alpha_2}",
        )

        assert base.La.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.La,
            base.SIEGBAHN,
            "La",
            "L\u03b1",
            "L&alpha;",
            "\\ensuremath{\\mathrm{L}\\alpha}",
        )
        yield XrayTransitionNotation(
            JENKINS1991,
            base.La,
            base.IUPAC,
            "L3-M(4,5)",
            "L3\u2013M(4,5)",
            "L<sub>3</sub>&ndash;M<sub>4,5</sub>",
            "\\ensuremath{\\mathrm{L}_3}--\\ensuremath{\\mathrm{M}_{4,5}}",
        )

        assert base.Lb1.source_subshell == base.M4
        assert base.Lb1.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb1,
            base.SIEGBAHN,
            "Lb1",
            "L\u03b21",
            "L&beta;<sub>1</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_1}",
        )

        assert base.Lb2.source_subshell == base.N5
        assert base.Lb2.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb2,
            base.SIEGBAHN,
            "Lb2",
            "L\u03b22",
            "L&beta;<sub>2</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_2}",
        )

        assert base.Lb3.source_subshell == base.M3
        assert base.Lb3.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb3,
            base.SIEGBAHN,
            "Lb3",
            "L\u03b23",
            "L&beta;<sub>3</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_3}",
        )

        assert base.Lb4.source_subshell == base.M2
        assert base.Lb4.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb4,
            base.SIEGBAHN,
            "Lb4",
            "L\u03b24",
            "L&beta;<sub>4</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_4}",
        )

        assert base.Lb3_4.destination_subshell == base.L1
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lb3_4,
            base.SIEGBAHN,
            "Lb3,4",
            "L\u03b23,4",
            "L&beta;<sub>3,4</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_{3,4}}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lb3_4,
            base.IUPAC,
            "L1-M(2,3)",
            "L1\u2013M(2,3)",
            "L<sub>1</sub>&ndash;M<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{L}_1}--\\ensuremath{\\mathrm{M}_{2,3}}",
        )

        assert base.Lb5_1.source_subshell == base.O5
        assert base.Lb5_1.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lb5_1,
            base.SIEGBAHN,
            "Lb5I",
            "L\u03b25I",
            "L&beta;<sub>5</sub><sup>I</sup>",
            "\\ensuremath{\\mathrm{L}\\beta_5^I}",
        )

        assert base.Lb5_2.source_subshell == base.O4
        assert base.Lb5_2.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lb5_2,
            base.SIEGBAHN,
            "Lb5II",
            "L\u03b25II",
            "L&beta;<sub>5</sub><sup>II</sup>",
            "\\ensuremath{\\mathrm{L}\\beta_5^{II}}",
        )

        assert base.Lb5.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lb5,
            base.SIEGBAHN,
            "Lb5",
            "L\u03b25",
            "L&beta;<sub>5</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_5}",
        )

        assert base.Lb6.source_subshell == base.N1
        assert base.Lb6.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb6,
            base.SIEGBAHN,
            "Lb6",
            "L\u03b26",
            "L&beta;<sub>6</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_6}",
        )

        assert base.Lb7.source_subshell == base.O1
        assert base.Lb7.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb7,
            base.SIEGBAHN,
            "Lb7",
            "L\u03b27",
            "L&beta;<sub>7</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_7}",
        )

        assert base.Lb9.source_subshell == base.M5
        assert base.Lb9.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb9,
            base.SIEGBAHN,
            "Lb9",
            "L\u03b29",
            "L&beta;<sub>9</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_9}",
        )

        assert base.Lb10.source_subshell == base.M4
        assert base.Lb10.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb10,
            base.SIEGBAHN,
            "Lb10",
            "L\u03b210",
            "L&beta;<sub>10</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_{10}}",
        )

        assert base.Lb15.source_subshell == base.N4
        assert base.Lb15.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb15,
            base.SIEGBAHN,
            "Lb15",
            "L\u03b215",
            "L&beta;<sub>15</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_{15}}",
        )

        assert base.Lb2_15.destination_subshell == base.L3
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lb2_15,
            base.SIEGBAHN,
            "Lb2,15",
            "L\u03b22,15",
            "L&beta;<sub>2,15</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_{2,15}}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lb2_15,
            base.IUPAC,
            "L3-N(4,5)",
            "L3\u2013N(4,5)",
            "L<sub>3</sub>&ndash;N<sub>4,5</sub>",
            "\\ensuremath{\\mathrm{L}_3}--\\ensuremath{\\mathrm{N}_{4,5}}",
        )

        assert base.Lb17.source_subshell == base.M3
        assert base.Lb17.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lb17,
            base.SIEGBAHN,
            "Lb17",
            "L\u03b217",
            "L&beta;<sub>17</sub>",
            "\\ensuremath{\\mathrm{L}\\beta_{17}}",
        )

        assert base.Lg1.source_subshell == base.N4
        assert base.Lg1.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg1,
            base.SIEGBAHN,
            "Lg1",
            "L\u03b31",
            "L&gamma;<sub>1</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_1}",
        )

        assert base.Lg2.source_subshell == base.N2
        assert base.Lg2.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg2,
            base.SIEGBAHN,
            "Lg2",
            "L\u03b32",
            "L&gamma;<sub>2</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_2}",
        )

        assert base.Lg3.source_subshell == base.N3
        assert base.Lg3.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg3,
            base.SIEGBAHN,
            "Lg3",
            "L\u03b33",
            "L&gamma;<sub>3</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_3}",
        )

        assert base.Lg2_3.destination_subshell == base.L1
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lg2_3,
            base.SIEGBAHN,
            "Lg2,3",
            "L\u03b32,3",
            "L&gamma;<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_{2,3}}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Lg2_3,
            base.IUPAC,
            "L1-N(2,3)",
            "L1\u2013N(2,3)",
            "L<sub>1</sub>&ndash;N<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{L}_1}--\\ensuremath{\\mathrm{N}_{2,3}}",
        )

        assert base.Lg4.source_subshell == base.O3
        assert base.Lg4.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg4,
            base.SIEGBAHN,
            "Lg4",
            "L\u03b34",
            "L&gamma;<sub>4</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_4}",
        )

        assert base.Lg4p.source_subshell == base.O2
        assert base.Lg4p.destination_subshell == base.L1
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg4p,
            base.SIEGBAHN,
            "Lg4'",
            "L\u03b34\u2032",
            "L&gamma;<sub>4&prime;</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_{4\\prime}}",
        )

        assert base.Lg5.source_subshell == base.N1
        assert base.Lg5.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg5,
            base.SIEGBAHN,
            "Lg5",
            "L\u03b35",
            "L&gamma;<sub>5</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_5}",
        )

        assert base.Lg6.source_subshell == base.O4
        assert base.Lg6.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg6,
            base.SIEGBAHN,
            "Lg6",
            "L\u03b36",
            "L&gamma;<sub>6</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_6}",
        )

        assert base.Lg8.source_subshell == base.O1
        assert base.Lg8.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lg8,
            base.SIEGBAHN,
            "Lg8",
            "L\u03b38",
            "L&gamma;<sub>8</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_8}",
        )

        assert base.Lg11.source_subshell == base.N5
        assert base.Lg11.destination_subshell == base.L1
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lg11,
            base.SIEGBAHN,
            "Lg11",
            "L\u03b311",
            "L&gamma;<sub>11</sub>",
            "\\ensuremath{\\mathrm{L}\\gamma_11}",
        )

        assert base.Ln.source_subshell == base.M1
        assert base.Ln.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ln,
            base.SIEGBAHN,
            "Ln",
            "L\u03b7",
            "L&eta;",
            "\\ensuremath{\\mathrm{L}\\eta}",
        )

        assert base.Ll.source_subshell == base.M1
        assert base.Ll.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ll,
            base.SIEGBAHN,
            "Ll",
            "L\u2113",
            "L&#x2113;",
            "\\ensuremath{\\mathrm{L}\\ell}",
        )

        assert base.Ll_n.source_subshell == base.M1
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Ll_n,
            base.SIEGBAHN,
            "Ll,n",
            "L\u2113,\u03b7",
            "L&#x2113;,&eta;",
            "\\ensuremath{\\mathrm{L}\\ell,\\eta}",
        )
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.Ll_n,
            base.IUPAC,
            "L(2,3)-M1",
            "L(2,3)\u2013M1",
            "L<sup>2,3</sup>&ndash;M<sub>1</sub>",
            "\\ensuremath{\\mathrm{L}_{2,3}}--\\ensuremath{\\mathrm{M}_1}",
        )

        assert base.Ls.source_subshell == base.M3
        assert base.Ls.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ls,
            base.SIEGBAHN,
            "Ls",
            "Ls",
            "Ls",
            "\\ensuremath{\\mathrm{L}s}",
        )

        assert base.Lt.source_subshell == base.M2
        assert base.Lt.destination_subshell == base.L3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lt,
            base.SIEGBAHN,
            "Lt",
            "Lt",
            "Lt",
            "\\ensuremath{\\mathrm{L}t}",
        )

        assert base.Lu_1.source_subshell == base.N7
        assert base.Lu_1.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lu_1,
            base.SIEGBAHN,
            "LuI",
            "LuI",
            "Lu<sup>I</sup>",
            "\\ensuremath{\\mathrm{L}u^I}",
        )

        assert base.Lu_2.source_subshell == base.N6
        assert base.Lu_2.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lu_2,
            base.SIEGBAHN,
            "LuII",
            "LuII",
            "Lu<sup>II</sup>",
            "\\ensuremath{\\mathrm{L}u^{II}}",
        )

        assert base.Lu.destination_subshell == base.L3
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Lu,
            base.SIEGBAHN,
            "Lu",
            "Lu",
            "Lu",
            "\\ensuremath{\\mathrm{L}u}",
        )

        assert base.Lv.source_subshell == base.N6
        assert base.Lv.destination_subshell == base.L2
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Lv,
            base.SIEGBAHN,
            "Lv",
            "L\u03bd",
            "L&nu;",
            "\\ensuremath{\\mathrm{L}\\nu}",
        )

        assert base.Ma1.source_subshell == base.N7
        assert base.Ma1.destination_subshell == base.M5
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ma1,
            base.SIEGBAHN,
            "Ma1",
            "M\u03b11",
            "M&alpha;<sub>1</sub>",
            "\\ensuremath{\\mathrm{M}\\alpha_1}",
        )

        assert base.Ma2.source_subshell == base.N6
        assert base.Ma2.destination_subshell == base.M5
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ma2,
            base.SIEGBAHN,
            "Ma2",
            "M\u03b12",
            "M&alpha;<sub>2</sub>",
            "\\ensuremath{\\mathrm{M}\\alpha_2}",
        )

        assert base.Ma.destination_subshell == base.M5
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ma,
            base.SIEGBAHN,
            "Ma",
            "M\u03b1",
            "M&alpha;",
            "\\ensuremath{\\mathrm{M}\\alpha}",
        )
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Ma,
            base.IUPAC,
            "M5-N(6,7)",
            "M5\u2013N(6,7)",
            "M<sub>5</sub>&ndash;N<sub>6,7</sub>",
            "\\ensuremath{\\mathrm{M}_5}--\\ensuremath{\\mathrm{N}_{6,7}}",
        )

        assert base.Mb.source_subshell == base.N6
        assert base.Mb.destination_subshell == base.M4
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Mb,
            base.SIEGBAHN,
            "Mb",
            "M\u03b2",
            "M&beta;",
            "\\ensuremath{\\mathrm{M}\\beta}",
        )

        assert base.Mg.source_subshell == base.N5
        assert base.Mg.destination_subshell == base.M3
        yield XrayTransitionNotation(
            JENKINS1991,
            base.Mg,
            base.SIEGBAHN,
            "Mg",
            "M\u03b3",
            "M&gamma;",
            "\\ensuremath{\\mathrm{M}\\gamma}",
        )

        assert base.Mz1.source_subshell == base.N3
        assert base.Mz1.destination_subshell == base.M5
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Mz1,
            base.SIEGBAHN,
            "Mz1",
            "M\u03b61",
            "M&zeta;<sub>1</sub>",
            "\\ensuremath{\\mathrm{M}\\zeta_1}",
        )

        assert base.Mz2.source_subshell == base.N2
        assert base.Mz2.destination_subshell == base.M4
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Mz2,
            base.SIEGBAHN,
            "Mz2",
            "M\u03b62",
            "M&zeta;<sub>2</sub>",
            "\\ensuremath{\\mathrm{M}\\zeta_2}",
        )

        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Mz,
            base.SIEGBAHN,
            "Mz",
            "M\u03b6",
            "M&zeta;",
            "\\ensuremath{\\mathrm{M}\\zeta}",
        )
        yield XrayTransitionNotation(
            BEARDEN1967,
            base.Mz,
            base.IUPAC,
            "M(4,5)-N(2,3)",
            "M(4,5)\u2013N(2,3)",
            "M<sub>4,5</sub>&ndash;N<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{M}_{4,5}}--\\ensuremath{\\mathrm{N}_{2,3}}",
        )

        assert base.M1N2_3.destination_subshell == base.M1
        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.M1N2_3,
            base.IUPAC,
            "M1-N(2,3)",
            "M1\u2013N(2,3)",
            "M<sub>1</sub>&ndash;N<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{M}_1}--\\ensuremath{\\mathrm{N}_{2,3}}",
        )

        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.M2_3M4_5,
            base.IUPAC,
            "M(2,3)-M(4,5)",
            "M(2,3)\u2013M(4,5)",
            "M<sub>2,3</sub>&ndash;M<sub>4,5</sub>",
            "\\ensuremath{\\mathrm{M}_{2,3}}--\\ensuremath{\\mathrm{M}_{4,5}}",
        )

        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.M4_5O2_3,
            base.IUPAC,
            "M(4,5)-O(2,3)",
            "M(4,5)\u2013O(2,3)",
            "M<sub>4,5</sub>&ndash;O<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{M}_{4,5}}--\\ensuremath{\\mathrm{O}_{2,3}}",
        )

        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.M3O4_5,
            base.IUPAC,
            "M3-O(4,5)",
            "M3\u2013O(4,5)",
            "M<sub>3</sub>&ndash;O<sub>4,5</sub>",
            "\\ensuremath{\\mathrm{M}_3}--\\ensuremath{\\mathrm{O}_{4,5}}",
        )

        yield XrayTransitionNotation(
            base.UNATTRIBUTED,
            base.M4O2_3,
            base.IUPAC,
            "M4-O(2,3)",
            "M4\u2013O(2,3)",
            "M<sub>4</sub>&ndash;O<sub>2,3</sub>",
            "\\ensuremath{\\mathrm{M}_4}--\\ensuremath{\\mathrm{O}_{2,3}}",
        )


class SeriesXrayTransitionNotationParser(base._Parser):
    def __iter__(self):
        for n in range(1, MAX_N + 1):
            transition = XrayTransition(
                None, None, None, n, None, None
            )  # All transitions to this shell

            (
                ascii,
                utf16,
                html,
                latex,
            ) = AtomicShellNotationParser._create_entry_siegbahn(n)
            yield XrayTransitionNotation(
                base.UNATTRIBUTED, transition, base.SIEGBAHN, ascii, utf16, html, latex
            )

            ascii, utf16, html, latex = AtomicShellNotationParser._create_entry_iupac(n)
            yield XrayTransitionNotation(
                base.UNATTRIBUTED, transition, base.IUPAC, ascii, utf16, html, latex
            )


class FamilyXrayTransitionNotationParser(base._Parser):
    def __iter__(self):
        families = set()
        for transition, _src_i, dst_i in iter_transitions(MAX_N):
            dst = transition.destination_subshell
            families.add((dst, dst_i))

        # Create family notations
        for atomic_subshell, i in families:
            if i == 0:  # Skip K, already a series
                continue

            n = atomic_subshell.n
            l = atomic_subshell.l
            j_n = atomic_subshell.j_n

            transition = XrayTransition(
                None, None, None, n, l, j_n
            )  # All transitions to this subshell

            (
                ascii,
                utf16,
                html,
                latex,
            ) = AtomicSubshellNotationParser._create_entry_siegbahn(n, l, j_n, i)
            yield XrayTransitionNotation(
                base.UNATTRIBUTED, transition, base.SIEGBAHN, ascii, utf16, html, latex
            )

            (
                ascii,
                utf16,
                html,
                latex,
            ) = AtomicSubshellNotationParser._create_entry_iupac(n, l, j_n, i)
            yield XrayTransitionNotation(
                base.UNATTRIBUTED, transition, base.IUPAC, ascii, utf16, html, latex
            )



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.cbook import ProgressMixin, ProgressReportMixin

# Globals and constants variables.


class MockProgress(ProgressMixin):
    pass


class MockProgressReport(ProgressReportMixin):
    pass


@pytest.fixture
def progress():
    return MockProgress()


def test_progress_update(progress):
    progress.update(50)
    assert progress.progress == 50


@pytest.fixture
def progress_report(progress):
    report = MockProgressReport()
    report.add_reporthook(lambda p: progress.update(p))
    return report


def test_progress_report_update(progress_report, progress):
    assert progress.progress == 0

    progress_report.update(50)

    assert progress_report.progress == 50
    assert progress.progress == 50



"""
Parsers from data collected from Wikipedia.
"""

# Standard library modules.
import os
import logging

logger = logging.getLogger(__name__)

# Third party modules.
import requests

# Local modules.
import pyxray.parser.base as base
from pyxray.descriptor import Reference, Element, Language
from pyxray.property import ElementName

# Globals and constants variables.

WIKIPEDIA = Reference("wikipedia", author="Wikipedia contributors")


class WikipediaElementNameParser(base._Parser):

    NAMES_EN = [
        "Hydrogen",
        "Helium",
        "Lithium",
        "Beryllium",
        "Boron",
        "Carbon",
        "Nitrogen",
        "Oxygen",
        "Fluorine",
        "Neon",
        "Sodium",
        "Magnesium",
        "Aluminium",
        "Silicon",
        "Phosphorus",
        "Sulfur",
        "Chlorine",
        "Argon",
        "Potassium",
        "Calcium",
        "Scandium",
        "Titanium",
        "Vanadium",
        "Chromium",
        "Manganese",
        "Iron",
        "Cobalt",
        "Nickel",
        "Copper",
        "Zinc",
        "Gallium",
        "Germanium",
        "Arsenic",
        "Selenium",
        "Bromine",
        "Krypton",
        "Rubidium",
        "Strontium",
        "Yttrium",
        "Zirconium",
        "Niobium",
        "Molybdenum",
        "Technetium",
        "Ruthenium",
        "Rhodium",
        "Palladium",
        "Silver",
        "Cadmium",
        "Indium",
        "Tin",
        "Antimony",
        "Tellurium",
        "Iodine",
        "Xenon",
        "Cesium",
        "Barium",
        "Lanthanum",
        "Cerium",
        "Praseodymium",
        "Neodymium",
        "Promethium",
        "Samarium",
        "Europium",
        "Gadolinium",
        "Terbium",
        "Dysprosium",
        "Holmium",
        "Erbium",
        "Thulium",
        "Ytterbium",
        "Lutetium",
        "Hafnium",
        "Tantalum",
        "Tungsten",
        "Rhenium",
        "Osmium",
        "Iridium",
        "Platinum",
        "Gold",
        "Mercury",
        "Thallium",
        "Lead",
        "Bismuth",
        "Polonium",
        "Astatine",
        "Radon",
        "Francium",
        "Radium",
        "Actinium",
        "Thorium",
        "Protactinium",
        "Uranium",
        "Neptunium",
        "Plutonium",
        "Americium",
        "Curium",
        "Berkelium",
        "Californium",
        "Einsteinium",
        "Fermium",
        "Mendelevium",
        "Nobelium",
        "Lawrencium",
        "Rutherfordium",
        "Dubnium",
        "Seaborgium",
        "Bohrium",
        "Hassium",
        "Meitnerium",
        "Darmstadtium",
        "Roentgenium",
        "Copernicium",
        "Ununtrium",
        "Flerovium",
        "Ununpentium",
        "Livermorium",
        "Ununseptium",
        "Ununoctium",
    ]

    # 30 most spoken languages in the world
    # https://en.wikipedia.org/wiki/List_of_languages_by_number_of_native_speakers
    LANGUAGES = [
        "cmn",
        "es",
        "en",
        "hi",
        "ar",
        "pt",
        "bn",
        "ru",
        "ja",
        "pa",
        "de",
        "jv",
        "wuu",
        "ms",
        "te",
        "vi",
        "ko",
        "fr",
        "mr",
        "ta",
        "ur",
        "tr",
        "it",
        "yue",
        "th",
        "gu",
        "cjy",
        "nan",
        "fa",
        "pl",
    ]

    def _find_wikipedia_names(self, name_en):
        """
        Finds all Wikipedia pages referring to the specified name in English and
        returns a dictionary where the keys are the language code and the values
        are the titles of the corresponding pages.
        """
        url = "https://en.wikipedia.org/w/api.php"
        params = {
            "action": "query",
            "titles": name_en,
            "prop": "langlinks",
            "lllimit": 500,
            "format": "json",
        }
        r = requests.get(url, params=params)
        if not r:
            raise ValueError("Could not find wikipedia page: {0}".format(name_en))
        out = r.json()

        names = {}
        pages = out["query"]["pages"]
        for page in pages:
            for langlink in pages[page].get("langlinks", []):
                names[langlink["lang"]] = langlink["*"]

        return names

    def __iter__(self):
        length = len(self.NAMES_EN)
        for z, name_en in enumerate(self.NAMES_EN, 1):
            element = Element(z)
            language = Language("en")
            prop = ElementName(WIKIPEDIA, element, language, name_en)
            logger.debug("Parsed: {0}".format(prop))
            yield prop

            names = self._find_wikipedia_names(name_en)
            for code, name in names.items():
                if code not in self.LANGUAGES:
                    continue
                language = Language(code)
                prop = ElementName(WIKIPEDIA, element, language, name)
                logger.debug("Parsed: {0}".format(prop))
                yield prop

            self.update(int((z - 1) / length * 100.0))



""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pyxray.parser.notation import (
    iter_subshells,
    ElementSymbolParser,
    AtomicShellNotationParser,
    AtomicSubshellNotationParser,
    GenericXrayTransitionNotationParser,
    KnownXrayTransitionNotationParser,
    SeriesXrayTransitionNotationParser,
    FamilyXrayTransitionNotationParser,
)

# Globals and constants variables.


def test_iter_subshells():
    assert len(list(iter_subshells(1))) == 1
    assert len(list(iter_subshells(2))) == 4
    assert len(list(iter_subshells(7))) == 49


def test_unattributed_symbol():
    parser = ElementSymbolParser()
    assert len(list(parser)) == 118


def test_atomicshell_notation():
    parser = AtomicShellNotationParser()
    assert len(list(parser)) == 21


def test_atomicsubshell_notation():
    parser = AtomicSubshellNotationParser()
    assert len(list(parser)) == 147


def test_generic_xray_transition_notation():
    parser = GenericXrayTransitionNotationParser()
    assert len(list(parser)) == 1176


def test_known_xray_transition_notation():
    parser = KnownXrayTransitionNotationParser()
    assert len(list(parser)) == 76


def test_series_xray_transition_notation():
    parser = SeriesXrayTransitionNotationParser()
    assert len(list(parser)) == 14


def test_family_xray_transition_notation():
    parser = FamilyXrayTransitionNotationParser()
    assert len(list(parser)) == 96



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.

# Local modules.
from pyxray.parser.perkins1991 import Perkins1991Parser

# Globals and constants variables.


def test_perkins1991():
    parser = Perkins1991Parser()
    assert len(list(parser)) == 19189



#!/usr/bin/env python
""" """

# Standard library modules.
import dataclasses

# Third party modules.
import pytest

# Local modules.
from pyxray.descriptor import (
    Element,
    AtomicShell,
    AtomicSubshell,
    Reference,
    XrayLine,
    XrayTransition,
    Language,
    Notation,
)

# Globals and constants variables.


@pytest.fixture
def element():
    return Element(6)


def test_element(element):
    assert element.z == 6
    assert element.atomic_number == 6


def test_element_eq(element):
    assert element == Element(6)


def test_element_hash(element):
    assert hash(element) == hash(Element(6))


def test_element_repr(element):
    assert repr(element) == "Element(z=6)"


def test_element_validate():
    with pytest.raises(ValueError):
        Element(0)

    with pytest.raises(ValueError):
        Element(119)


def test_element_frozen(element):
    with pytest.raises(dataclasses.FrozenInstanceError):
        element.atomic_number = 7

    with pytest.raises(dataclasses.FrozenInstanceError):
        del element.atomic_number

    with pytest.raises(dataclasses.FrozenInstanceError):
        element.abc = 7


@pytest.fixture
def atomicshell():
    return AtomicShell(3)


def test_atomicshell(atomicshell):
    assert atomicshell.n == 3
    assert atomicshell.principal_quantum_number == 3


def test_atomicshell_eq(atomicshell):
    assert atomicshell == AtomicShell(3)


def test_atomicshell_hash(atomicshell):
    assert hash(atomicshell) == hash(AtomicShell(3))


def test_atomicshell_repr(atomicshell):
    assert repr(atomicshell) == "AtomicShell(n=3)"


def test_atomicshell_validable():
    with pytest.raises(ValueError):
        AtomicShell(0)


def test_atomicshell_frozen(atomicshell):
    with pytest.raises(dataclasses.FrozenInstanceError):
        atomicshell.n = 7

    with pytest.raises(dataclasses.FrozenInstanceError):
        del atomicshell.n

    with pytest.raises(dataclasses.FrozenInstanceError):
        atomicshell.abc = 7


@pytest.fixture
def atomicsubshell():
    return AtomicSubshell(3, 0, 1)


def test_atomicsubshell(atomicsubshell):
    assert atomicsubshell.n == 3
    assert atomicsubshell.atomic_shell == AtomicShell(3)
    assert atomicsubshell.principal_quantum_number == 3
    assert atomicsubshell.l == 0
    assert atomicsubshell.azimuthal_quantum_number == 0
    assert atomicsubshell.j_n == 1
    assert atomicsubshell.total_angular_momentum_nominator == 1
    assert atomicsubshell.j == 0.5
    assert atomicsubshell.total_angular_momentum == 0.5


def test_atomicsubshell_eq(atomicsubshell):
    assert atomicsubshell == AtomicSubshell(3, 0, 1)
    assert atomicsubshell == AtomicSubshell(AtomicShell(3), 0, 1)


def test_atomicsubshell_hash(atomicsubshell):
    assert hash(atomicsubshell) == hash(AtomicSubshell(3, 0, 1))
    assert hash(atomicsubshell) == hash(AtomicSubshell(AtomicShell(3), 0, 1))


def test_atomicsubshell_repr(atomicsubshell):
    assert repr(atomicsubshell) == "AtomicSubshell(n=3, l=0, j=0.5)"


def test_atomicsubshell_validate():
    with pytest.raises(ValueError):
        AtomicSubshell(3, 0, 5)

    with pytest.raises(ValueError):
        AtomicSubshell(3, -1, 1)

    with pytest.raises(ValueError):
        AtomicSubshell(3, 3, 1)


def test_atomicsubshell_frozen(atomicsubshell):
    with pytest.raises(dataclasses.FrozenInstanceError):
        atomicsubshell.n = 7

    with pytest.raises(dataclasses.FrozenInstanceError):
        del atomicsubshell.n

    with pytest.raises(dataclasses.FrozenInstanceError):
        atomicsubshell.abc = 7


@pytest.fixture(
    params=[
        (AtomicSubshell(2, 0, 1), AtomicSubshell(1, 0, 1)),
        (2, 0, 1, 1, 0, 1),
        ((2, 0, 1), (1, 0, 1)),
        (AtomicSubshell(2, 0, 1), 1, 0, 1),
        (2, 0, 1, AtomicSubshell(1, 0, 1)),
    ]
)
def xraytransition(request):
    return XrayTransition(*request.param)


def test_xraytransition(xraytransition):
    assert xraytransition.source_principal_quantum_number == 2
    assert xraytransition.source_azimuthal_quantum_number == 0
    assert xraytransition.source_total_angular_momentum_nominator == 1
    assert xraytransition.destination_principal_quantum_number == 1
    assert xraytransition.destination_azimuthal_quantum_number == 0
    assert xraytransition.destination_total_angular_momentum_nominator == 1

    assert xraytransition.source_subshell.n == 2
    assert xraytransition.source_subshell.l == 0
    assert xraytransition.source_subshell.j_n == 1
    assert xraytransition.destination_subshell.n == 1
    assert xraytransition.destination_subshell.l == 0
    assert xraytransition.destination_subshell.j_n == 1


def test_xraytransition_eq(xraytransition):
    assert xraytransition == XrayTransition((2, 0, 1), (1, 0, 1))


def test_xraytransition_hash(xraytransition):
    assert hash(xraytransition) == hash(XrayTransition((2, 0, 1), (1, 0, 1)))
    assert hash(xraytransition) == hash(
        XrayTransition(AtomicSubshell(2, 0, 1), AtomicSubshell(1, 0, 1))
    )


def test_xraytransition_repr(xraytransition):
    assert (
        repr(xraytransition) == "XrayTransition([n=2, l=0, j=0.5] -> [n=1, l=0, j=0.5])"
    )


def test_xraytransition_frozen(xraytransition):
    with pytest.raises(dataclasses.FrozenInstanceError):
        xraytransition.source_subshell = AtomicSubshell(2, 0, 1)

    with pytest.raises(dataclasses.FrozenInstanceError):
        del xraytransition.source_subshell

    with pytest.raises(dataclasses.FrozenInstanceError):
        xraytransition.abc = 7


@pytest.fixture
def xraytransitionset():
    return XrayTransition(
        2, 1, None, 1, 0, 1
    )  # L2 (2, 1, 0.5) and L3 (2, 1, 1.5) to K (1, 0, 0.5)


def test_xraytransitionset(xraytransitionset):
    assert xraytransitionset.source_subshell.n == 2
    assert xraytransitionset.source_subshell.l == 1
    assert xraytransitionset.source_subshell.j_n == None
    assert xraytransitionset.destination_subshell.n == 1
    assert xraytransitionset.destination_subshell.l == 0
    assert xraytransitionset.destination_subshell.j_n == 1


def test_xraytransitionset_repr(xraytransitionset):
    assert (
        repr(xraytransitionset)
        == "XrayTransition([n=2, l=1, j=*] -> [n=1, l=0, j=0.5])"
    )


@pytest.fixture
def xrayline():
    return XrayLine(
        Element(118), XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1, 0.2, 0.3
    )


def test_xrayline(xrayline):
    assert xrayline.element.z == 118
    assert xrayline.transition == XrayTransition(2, 0, 1, 1, 0, 1)
    assert xrayline.iupac == "a"
    assert xrayline.siegbahn == "b"
    assert xrayline.energy_eV == pytest.approx(0.1, abs=1e-4)
    assert xrayline.probability == pytest.approx(0.2, abs=1e-4)
    assert xrayline.relative_weight == pytest.approx(0.3, abs=1e-4)


def test_xrayline_eq(xrayline):
    assert xrayline == XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1)
    assert xrayline == XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "z", "b", 0.1)
    assert xrayline == XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "z", 0.1)
    assert xrayline == XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 99.0)
    assert xrayline == XrayLine(
        118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1, 99.0
    )
    assert xrayline == XrayLine(
        118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1, 0.2, 99.0
    )
    assert xrayline == XrayLine(
        118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1, 0.2, 0.3
    )

    assert xrayline != XrayLine(117, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1)
    assert xrayline != XrayLine(118, XrayTransition(3, 0, 1, 1, 0, 1), "a", "b", 0.1)


@pytest.mark.parametrize(
    "other",
    [
        XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.1),
        XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "z", "b", 0.1),
        XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "z", 0.1),
        XrayLine(118, XrayTransition(2, 0, 1, 1, 0, 1), "a", "b", 0.2),
    ],
)
def test_xrayline_hash(xrayline, other):
    assert hash(xrayline) == hash(other)


def test_xrayline_repr(xrayline):
    assert repr(xrayline) == "XrayLine(a)"


def test_xrayline_frozen(xrayline):
    with pytest.raises(dataclasses.FrozenInstanceError):
        xrayline.element = 117

    with pytest.raises(dataclasses.FrozenInstanceError):
        del xrayline.element

    with pytest.raises(dataclasses.FrozenInstanceError):
        xrayline.abc = 7


@pytest.fixture
def language():
    return Language("en")


def test_language(language):
    assert language.key == "en"


def test_language_eq(language):
    assert language == Language("en")
    assert language == Language("EN")
    assert language != Language("fr")


def test_language_hash(language):
    assert hash(language) == hash(Language("EN"))


def test_language_repr(language):
    assert repr(language) == "Language(en)"


def test_language_validate():
    with pytest.raises(ValueError):
        Language("english")

    with pytest.raises(ValueError):
        Language("e")


def test_language_frozen(language):
    with pytest.raises(dataclasses.FrozenInstanceError):
        language.key = "fr"

    with pytest.raises(dataclasses.FrozenInstanceError):
        del language.key

    with pytest.raises(dataclasses.FrozenInstanceError):
        language.abc = 7


@pytest.fixture
def notation():
    return Notation("foo")


def test_notation(notation):
    assert notation.key == "foo"


def test_notation_eq(notation):
    assert notation == Notation("foo")
    assert notation == Notation("FOO")
    assert notation != Notation("bar")


def test_notation_hash(notation):
    assert hash(notation) == hash(Notation("FOO"))


def test_notation_repr(notation):
    assert repr(notation) == "Notation(foo)"


def test_notation_validate():
    with pytest.raises(ValueError):
        Notation("")


def test_notation_frozen(notation):
    with pytest.raises(dataclasses.FrozenInstanceError):
        notation.key = "bar"

    with pytest.raises(dataclasses.FrozenInstanceError):
        del notation.key

    with pytest.raises(dataclasses.FrozenInstanceError):
        notation.abc = 7


@pytest.fixture
def reference():
    return Reference("doe2016")


def test_reference(reference):
    assert reference.bibtexkey == "doe2016"


def test_reference_eq(reference):
    assert reference == Reference("doe2016")
    assert reference != Reference("doe2016", year=2016)


def test_reference_hash(reference):
    assert hash(reference) == hash(Reference("doe2016"))


def test_reference_repr(reference):
    assert repr(reference) == "Reference(doe2016)"


def test_reference_frozen(reference):
    with pytest.raises(dataclasses.FrozenInstanceError):
        reference.author = "bar"

    with pytest.raises(dataclasses.FrozenInstanceError):
        del reference.author

    with pytest.raises(dataclasses.FrozenInstanceError):
        reference.abc = 7



"""
Parsers from Campbell.
"""

# Standard library modules.
import logging
import os
import pkgutil

# Third party modules.

# Local modules.
import pyxray.parser.base as base
from pyxray.descriptor import Reference, Element
from pyxray.property import AtomicSubshellRadiativeWidth

# Globals and constants variables.
logger = logging.getLogger(__name__)

CAMPBELL2001 = Reference(
    "campbell2001",
    author="Campbell, J.L. and Papp, T.",
    year=2001,
    title="Widths of the atomic K-N7 levels",
    booktitle="Atomic Data and Nuclear Data Tables",
    pages="1-56",
    volume=77,
)

_SUBSHELL_LOOKUP = {
    "K": base.K,
    "L1": base.L1,
    "L2": base.L2,
    "L3": base.L3,
    "M1": base.M1,
    "M2": base.M2,
    "M3": base.M3,
    "M4": base.M4,
    "M5": base.M5,
    "N1": base.N1,
    "N2": base.N2,
    "N3": base.N3,
    "N4": base.N4,
    "N5": base.N5,
    "N6": base.N6,
    "N7": base.N7,
}

subshell_order = [
    "K",
    "L1",
    "L2",
    "L3",
    "M1",
    "M2",
    "M3",
    "M4",
    "M5",
    "N1",
    "N2",
    "N3",
    "N4",
    "N5",
    "N6",
    "N7",
]


class CampbellAtomicSubshellRadiativeWidthParser(base._Parser):
    def __iter__(self):
        relpath = os.path.join("..", "data", "campbell.asc")
        content = pkgutil.get_data(__name__, relpath).decode("utf8")

        shell_width = []
        for line in content.splitlines():
            line = line.strip()
            if not line:
                continue

            z = int(line[0:2])

            for s in range(1, 17):
                s_w = line[6 * s : 6 * s + 6].strip()
                if s_w == "":
                    continue
                shell_width.append([z, subshell_order[s - 1], float(s_w)])

        length = len(shell_width)
        for z, subshell, width in shell_width:
            if width is None:
                continue
            subshell = _SUBSHELL_LOOKUP[subshell]
            element = Element(z)
            prop = AtomicSubshellRadiativeWidth(CAMPBELL2001, element, subshell, width)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((z - 1) / length * 100.0))
            yield prop



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.

# Local modules.
from pyxray.parser.wikipedia import WikipediaElementNameParser

# Globals and constants variables.


def test_wikipedia():
    parser = WikipediaElementNameParser()
    assert len(list(parser)) > 0



# Standard library modules.
import os
import collections.abc
import inspect

# Third party modules.
import pkg_resources

# Local modules.
from pyxray.cbook import ProgressReportMixin
from pyxray.descriptor import AtomicSubshell, XrayTransition, Notation, Reference

# Globals and constants variables.
ENTRY_POINT = "pyxray.parser"

K = AtomicSubshell(1, 0, 1)
L1 = AtomicSubshell(2, 0, 1)
L2 = AtomicSubshell(2, 1, 1)
L3 = AtomicSubshell(2, 1, 3)
M1 = AtomicSubshell(3, 0, 1)
M2 = AtomicSubshell(3, 1, 1)
M3 = AtomicSubshell(3, 1, 3)
M4 = AtomicSubshell(3, 2, 3)
M5 = AtomicSubshell(3, 2, 5)
N1 = AtomicSubshell(4, 0, 1)
N2 = AtomicSubshell(4, 1, 1)
N3 = AtomicSubshell(4, 1, 3)
N4 = AtomicSubshell(4, 2, 3)
N5 = AtomicSubshell(4, 2, 5)
N6 = AtomicSubshell(4, 3, 5)
N7 = AtomicSubshell(4, 3, 7)
O1 = AtomicSubshell(5, 0, 1)
O2 = AtomicSubshell(5, 1, 1)
O3 = AtomicSubshell(5, 1, 3)
O4 = AtomicSubshell(5, 2, 3)
O5 = AtomicSubshell(5, 2, 5)
O6 = AtomicSubshell(5, 3, 5)
O7 = AtomicSubshell(5, 3, 7)
O8 = AtomicSubshell(5, 4, 7)
O9 = AtomicSubshell(5, 4, 9)
P1 = AtomicSubshell(6, 0, 1)
P2 = AtomicSubshell(6, 1, 1)
P3 = AtomicSubshell(6, 1, 3)
P4 = AtomicSubshell(6, 2, 3)
P5 = AtomicSubshell(6, 2, 5)
P6 = AtomicSubshell(6, 3, 5)
P7 = AtomicSubshell(6, 3, 7)
P8 = AtomicSubshell(6, 4, 7)
P9 = AtomicSubshell(6, 4, 9)
P10 = AtomicSubshell(6, 5, 9)
P11 = AtomicSubshell(6, 5, 11)
Q1 = AtomicSubshell(7, 0, 1)
Q2 = AtomicSubshell(7, 1, 1)
Q3 = AtomicSubshell(7, 1, 3)

Ka1 = XrayTransition(L3, K)
Ka2 = XrayTransition(L2, K)
Ka = XrayTransition(2, 1, None, K)
Kb1 = XrayTransition(M3, K)
Kb2_1 = XrayTransition(N3, K)
Kb2_2 = XrayTransition(N2, K)
Kb2 = XrayTransition(4, 1, None, K)
Kb3 = XrayTransition(M2, K)
Kb1_3 = XrayTransition(3, 1, None, K)  # K-M2,3
Kb4_1 = XrayTransition(N5, K)
Kb4_2 = XrayTransition(N4, K)
Kb4 = XrayTransition(4, 2, None, K)
Kb5_1 = XrayTransition(M5, K)
Kb5_2 = XrayTransition(M4, K)
Kb5 = XrayTransition(3, 2, None, K)  # K-M4,5
KO2_3 = XrayTransition(5, 1, None, K)  # K-O2,3

La1 = XrayTransition(M5, L3)
La2 = XrayTransition(M4, L3)
La = XrayTransition(3, 2, None, L3)
Lb1 = XrayTransition(M4, L2)
Lb2 = XrayTransition(N5, L3)
Lb3 = XrayTransition(M3, L1)
Lb4 = XrayTransition(M2, L1)
Lb3_4 = XrayTransition(3, 1, None, L1)  # L1-M2,3
Lb5_1 = XrayTransition(O5, L3)
Lb5_2 = XrayTransition(O4, L3)
Lb5 = XrayTransition(5, 2, None, L3)  # L3-O4,5
Lb6 = XrayTransition(N1, L3)
Lb7 = XrayTransition(O1, L3)
Lb9 = XrayTransition(M5, L1)
Lb10 = XrayTransition(M4, L1)
Lb15 = XrayTransition(N4, L3)
Lb2_15 = XrayTransition(4, 2, None, L3)  # L3-N4,5
Lb17 = XrayTransition(M3, L2)
Leta = XrayTransition(M1, L2)
Lg1 = XrayTransition(N4, L2)
Lg2 = XrayTransition(N2, L1)
Lg3 = XrayTransition(N3, L1)
Lg2_3 = XrayTransition(4, 1, None, L1)  # L1-N2,3
Lg4 = XrayTransition(O3, L1)
Lg4p = XrayTransition(O2, L1)
Lg5 = XrayTransition(N1, L2)
Lg6 = XrayTransition(O4, L2)
Lg8 = XrayTransition(O1, L2)

Lg11 = XrayTransition(N5, L1)
Ln = XrayTransition(M1, L2)
Ll = XrayTransition(M1, L3)
Ll_n = XrayTransition(M1, 2, 1, None)  # L2,3-M1
Ls = XrayTransition(M3, L3)
Lt = XrayTransition(M2, L3)
Lv = XrayTransition(N6, L2)
Lu_1 = XrayTransition(N7, L3)
Lu_2 = XrayTransition(N6, L3)
Lu = XrayTransition(4, 3, None, L3)

Ma1 = XrayTransition(N7, M5)
Ma2 = XrayTransition(N6, M5)
Ma = XrayTransition(4, 3, None, M5)
Mb = XrayTransition(N6, M4)
Mg = XrayTransition(N5, M3)
Mz1 = XrayTransition(N3, M5)
Mz2 = XrayTransition(N2, M4)
Mz = XrayTransition(4, 1, None, 3, 2, None)
M1N2_3 = XrayTransition(4, 1, None, M1)  # M1-N2,3
M2_3M4_5 = XrayTransition(3, 2, None, 3, 1, None)  # M2,3-M4,5
M4_5O2_3 = XrayTransition(5, 1, None, 3, 2, None)  # M4,5-O2,3
M3O4_5 = XrayTransition(5, 2, None, M3)  # M3-O4,5
M4O2_3 = XrayTransition(5, 1, None, M4)  # M4-O2,3

SIEGBAHN = Notation("siegbahn")
IUPAC = Notation("iupac")
ORBITAL = Notation("orbital")

UNATTRIBUTED = Reference("unattributed")

# Install cache
try:
    import requests_cache

    dirpath = os.path.join(os.path.dirname(__file__), "..", "data")
    filepath = os.path.join(dirpath, "cache")
    requests_cache.install_cache(filepath)
except ImportError:
    pass


class _Parser(collections.abc.Iterable, ProgressReportMixin):
    """
    (abstract) Class to parse X-ray related information from a data source.

    A parser is simply an iterable object which returns a new property at each
    iteration. For example::

    class ElementSymbolParser(_Parser):

        def __iter__(self):
            yield ElementSymbol(...)

    Each parser should be registered in the setup.py under the entry point:
    `pyxray.parser`
    """

    pass


def find_parsers():
    parsers = []
    for entry_point in pkg_resources.iter_entry_points(ENTRY_POINT):
        name = entry_point.name
        parser = entry_point.load()
        if inspect.isclass(parser):
            parser = parser()
        parsers.append((name, parser))
    return parsers



"""
Parsers from NIST.
"""

# Standard library modules.
import logging
import re
import os
import pkgutil

# Third party modules.

# Local modules.
import pyxray.parser.base as base
from pyxray.descriptor import Reference, Element
from pyxray.property import ElementAtomicWeight

# Globals and constants variables.
logger = logging.getLogger(__name__)

NIST = Reference(
    "coursey2015",
    author="J. S. Coursey, D. J. Schwab, J. J. Tsai, R. A. Dragoset",
    title="Atomic Weights and Isotopic Compositions",
    organization="NIST Physical Measurement Laboratory",
    year=2015,
)


class NISTElementAtomicWeightParser(base._Parser):
    def __iter__(self):
        relpath = os.path.join("..", "data", "nist_element_atomic_weight.html")
        content = pkgutil.get_data(__name__, relpath).decode("utf8")

        current_z = "1"
        value = 0

        atomic_weights = []
        for line in content.splitlines():
            z = re.search(r"Atomic\sNumber\s=\s([0-9]*)", line)
            relative_mass = re.search(
                r"Relative\sAtomic\sMass\s=\s([0-9]*.{0,1}[0-9]*)", line
            )
            composition = re.search(
                r"Isotopic\sComposition\s=\s([0-9]*.{0,1}[0-9]*)", line
            )
            if z != None:
                if current_z != z.group(1):
                    current_z = z.group(1)
                    if value == 0:
                        value = None
                    atomic_weights.append(value)
                    value = 0
            elif relative_mass != None:
                if relative_mass.group(1) == "":
                    r_m = 0.0
                else:
                    r_m = float(relative_mass.group(1))
            elif composition != None:
                if composition.group(1) == "":
                    c = 0.0
                else:
                    c = float(composition.group(1))
                value = value + r_m * c

        length = len(atomic_weights)
        for z, aw in enumerate(atomic_weights, 1):
            if aw is None:
                continue
            element = Element(z)
            prop = ElementAtomicWeight(NIST, element, aw)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((z - 1) / length * 100.0))
            yield prop



#!/usr/bin/env python
""" """

# Standard library modules.
import os
import tempfile
import shutil
import sqlite3

# Third party modules.
import pytest

# Local modules.
import pyxray.descriptor as descriptor
from pyxray.sql.data import SqlDatabase, NotFound
import pyxray.data

# Globals and constants variables.
K = descriptor.AtomicSubshell(1, 0, 1)
L3 = descriptor.AtomicSubshell(2, 1, 3)
L2 = descriptor.AtomicSubshell(2, 1, 1)


@pytest.fixture(scope="session")
def database(builder):
    return SqlDatabase(builder.engine)


@pytest.fixture
def database_real(tmp_path):
    return pyxray.data.database


def test_add_preferred_reference(database):
    database.clear_preferred_references()
    database.add_preferred_reference("lee1966")

    assert len(database.get_preferred_references()) == 1
    assert "lee1966" in database.get_preferred_references()

    database.clear_preferred_references()
    assert len(database.get_preferred_references()) == 0


def test_add_preferred_reference_not_found(database):
    with pytest.raises(NotFound):
        database.add_preferred_reference("foo")


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element(database, element):
    assert database.element(element) == descriptor.Element(118)


@pytest.mark.parametrize("reference", ["lee1966", "LEE1966"])
def test_reference(database, reference):
    assert database.element_name(118, "en", reference) == "Vibranium"


@pytest.mark.parametrize("element", ["Al", 13, "Aluminum"])
def test_element_notfound(database, element):
    with pytest.raises(NotFound):
        database.element(element)


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element_atomic_number(database, element):
    assert database.element_atomic_number(element) == 118


@pytest.mark.parametrize("element", ["Al", 13, "Aluminum"])
def test_element_atomic_number_notfound(database, element):
    with pytest.raises(NotFound):
        database.element_atomic_number(element)


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element_symbol(database, element):
    assert database.element_symbol(element) == "Vi"


@pytest.mark.parametrize("element", ["Al", 13, "Aluminum"])
def test_element_symbol_notfound(database, element):
    with pytest.raises(NotFound):
        database.element_symbol(element)


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element_name(database, element):
    assert database.element_name(element, "en") == "Vibranium"
    assert database.element_name(element, "en", "lee1966") == "Vibranium"


@pytest.mark.parametrize("element", ["Al", 13, "Aluminum"])
def test_element_name_notfound(database, element):
    with pytest.raises(NotFound):
        database.element_name(element, "en")


def test_element_name_notfound_wrong_language(database):
    with pytest.raises(NotFound):
        database.element_name(118, "fr")


def test_element_name_notfound_wrong_reference(database):
    with pytest.raises(NotFound):
        database.element_name(118, "en", "doe2016")


def test_element_atomic_weight_no_reference(database):
    assert database.element_atomic_weight(118) == pytest.approx(999.1, abs=1e-2)


def test_element_atomic_weight_lee1966(database):
    assert database.element_atomic_weight(118, "lee1966") == pytest.approx(
        999.1, abs=1e-2
    )


def test_element_atomic_weight_doe2016(database):
    assert database.element_atomic_weight(118, "doe2016") == pytest.approx(
        111.1, abs=1e-2
    )


def test_element_atomic_weight_preferred_reference(database):
    database.clear_preferred_references()
    database.add_preferred_reference("doe2016")
    assert database.element_atomic_weight(118) == pytest.approx(111.1, abs=1e-2)

    database.clear_preferred_references()
    database.add_preferred_reference("lee1966")
    assert database.element_atomic_weight(118) == pytest.approx(999.1, abs=1e-2)

    database.clear_preferred_references()


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element_mass_density_kg_per_m3(database, element):
    assert database.element_mass_density_kg_per_m3(element) == pytest.approx(
        999.2, abs=1e-2
    )


@pytest.mark.parametrize("element", [118, "Vi", "Vibranium"])
def test_element_mass_density_g_per_cm3(database, element):
    assert database.element_mass_density_g_per_cm3(element) == pytest.approx(
        0.9992, abs=1e-4
    )


@pytest.mark.parametrize("element,reference", [(118, None), (118, "lee1966")])
def test_element_xray_transitions(database, element, reference):
    transitions = database.element_xray_transitions(element, reference=reference)
    assert len(transitions) == 3

    assert descriptor.XrayTransition(L3, K) in transitions
    assert descriptor.XrayTransition(L2, K) in transitions
    assert descriptor.XrayTransition(2, 1, None, K) in transitions


@pytest.mark.parametrize(
    "xray_transition,expected",
    [
        (descriptor.XrayTransition(L3, K), 1),
        (descriptor.XrayTransition(2, 1, None, K), 2),
    ],
)
def test_element_xray_transitions_with_xray_transition(
    database, xray_transition, expected
):
    transitions = database.element_xray_transitions(118, xray_transition)
    assert len(transitions) == expected


@pytest.mark.parametrize(
    "element, expected", [(13, 14), (6, 2), (5, 3), (4, 3), (3, 2),]
)
def test_element_xray_transitions(database_real, element, expected):
    transitions = database_real.element_xray_transitions(element)
    assert len(transitions) == expected


@pytest.mark.parametrize("element,reference", [(118, "unknown"), (1, None)])
def test_element_xray_transitions_notfound(database, element, reference):
    with pytest.raises(NotFound):
        database.element_xray_transitions(element, reference)


def test_element_xray_transition(database):
    transition = database.element_xray_transition(118, "a")
    assert transition == descriptor.XrayTransition(L3, K)


@pytest.mark.parametrize("element,reference", [(1, "a"), (118, "g")])
def test_element_xray_transition_notfound(database, element, reference):
    with pytest.raises(NotFound):
        database.element_xray_transition(element, reference)


@pytest.mark.parametrize("atomic_shell", [1, "a", "b"])
def test_atomic_shell(database, atomic_shell):
    assert database.atomic_shell(atomic_shell) == descriptor.AtomicShell(1)


@pytest.mark.parametrize("atomic_shell", ["c", 3])
def test_atomic_shell_notfound(database, atomic_shell):
    with pytest.raises(NotFound):
        database.atomic_shell(atomic_shell)


@pytest.mark.parametrize(
    "encoding,expected", [("ascii", "a"), ("utf16", "b"), ("html", "c"), ("latex", "d")]
)
def test_atomic_shell_notation(database, encoding, expected):
    assert database.atomic_shell_notation(1, "mock", encoding) == expected


@pytest.mark.parametrize(
    "atomic_subshell", [descriptor.AtomicSubshell(1, 0, 1), (1, 0, 1), "a", "b"]
)
def test_atomic_subshell(database, atomic_subshell):
    assert database.atomic_subshell(atomic_subshell) == descriptor.AtomicSubshell(
        1, 0, 1
    )


@pytest.mark.parametrize("atomic_subshell", ["c", (3, 3, 3)])
def test_atomic_subshell_notfound(database, atomic_subshell):
    with pytest.raises(NotFound):
        database.atomic_subshell(atomic_subshell)


@pytest.mark.parametrize(
    "encoding,expected", [("ascii", "a"), ("utf16", "b"), ("html", "c"), ("latex", "d")]
)
def testatomic_subshell_notation(database, encoding, expected):
    assert (
        database.atomic_subshell_notation(
            descriptor.AtomicSubshell(1, 0, 1), "mock", encoding
        )
        == expected
    )


@pytest.mark.parametrize(
    "atomic_subshell", [descriptor.AtomicSubshell(1, 0, 1), (1, 0, 1), "a", "b"]
)
def test_atomic_subshell_binding_energy_eV(database, atomic_subshell):
    assert database.atomic_subshell_binding_energy_eV(
        118, atomic_subshell
    ) == pytest.approx(0.1, abs=1e-4)


@pytest.mark.parametrize(
    "atomic_subshell", [descriptor.AtomicSubshell(1, 0, 1), (1, 0, 1), "a", "b"]
)
def testatomic_subshell_radiative_width_eV(database, atomic_subshell):
    assert database.atomic_subshell_radiative_width_eV(
        118, atomic_subshell
    ) == pytest.approx(0.01, abs=1e-4)


@pytest.mark.parametrize(
    "atomic_subshell", [descriptor.AtomicSubshell(1, 0, 1), (1, 0, 1), "a", "b"]
)
def testatomic_subshell_nonradiative_width_eV(database, atomic_subshell):
    assert database.atomic_subshell_nonradiative_width_eV(
        118, atomic_subshell
    ) == pytest.approx(0.001, abs=1e-4)


@pytest.mark.parametrize(
    "atomic_subshell", [descriptor.AtomicSubshell(1, 0, 1), (1, 0, 1), "a", "b"]
)
def testatomic_subshell_occupancy(database, atomic_subshell):
    assert database.atomic_subshell_occupancy(118, atomic_subshell) == 1


@pytest.mark.parametrize(
    "xray_transition",
    ["a", ((2, 1, 3), (1, 0, 1)), (L3, K), descriptor.XrayTransition(L3, K)],
)
def test_xray_transition(database, xray_transition):
    assert database.xray_transition(xray_transition) == descriptor.XrayTransition(L3, K)


@pytest.mark.parametrize(
    "encoding,expected", [("ascii", "a"), ("utf16", "b"), ("html", "c"), ("latex", "d")]
)
def test_xray_transition_notation(database, encoding, expected):
    xray_transition = descriptor.XrayTransition(L3, K)
    assert (
        database.xray_transition_notation(xray_transition, "mock", encoding) == expected
    )


@pytest.mark.parametrize(
    "xray_transition",
    ["a", ((2, 1, 3), (1, 0, 1)), (L3, K), descriptor.XrayTransition(L3, K)],
)
def test_xray_transition_energy_eV_KL3(database, xray_transition):
    assert database.xray_transition_energy_eV(118, xray_transition) == pytest.approx(
        0.2, abs=1e-4
    )


@pytest.mark.parametrize(
    "xray_transition",
    ["e", ((2, 1, 1), (1, 0, 1)), (L2, K), descriptor.XrayTransition(L2, K)],
)
def test_xray_transition_energy_eV_KL2(database, xray_transition):
    assert database.xray_transition_energy_eV(118, xray_transition) == pytest.approx(
        0.4, abs=1e-4
    )


@pytest.mark.parametrize(
    "xray_transition",
    ["a", ((2, 1, 3), (1, 0, 1)), (L3, K), descriptor.XrayTransition(L3, K)],
)
def test_xray_transition_probability(database, xray_transition):
    assert database.xray_transition_probability(118, xray_transition) == pytest.approx(
        0.02, abs=1e-4
    )


@pytest.mark.parametrize(
    "xray_transition",
    ["a", ((2, 1, 3), (1, 0, 1)), (L3, K), descriptor.XrayTransition(L3, K)],
)
def test_xray_transition_relative_weight(database, xray_transition):
    assert database.xray_transition_relative_weight(
        118, xray_transition
    ) == pytest.approx(0.002, abs=1e-4)


def test_xray_line(database):
    xrayline = database.xray_line(118, "aa")

    assert xrayline.element.atomic_number == 118
    assert xrayline.iupac == "Vi bb"
    assert xrayline.siegbahn == "Vi bb"
    assert xrayline.energy_eV == pytest.approx(0.2, abs=1e-3)



"""
Current implementation of the database
"""

__all__ = [
    "get_preferred_references",
    "add_preferred_reference",
    "clear_preferred_references",
    "element",
    "element_atomic_number",
    "element_symbol",
    "element_name",
    "element_atomic_weight",
    "element_mass_density_kg_per_m3",
    "element_mass_density_g_per_cm3",
    "element_xray_transitions",
    "element_xray_transition",
    "print_element_xray_transitions",
    "atomic_shell",
    "atomic_shell_notation",
    "atomic_subshell",
    "atomic_subshell_notation",
    "atomic_subshell_binding_energy_eV",
    "atomic_subshell_radiative_width_eV",
    "atomic_subshell_nonradiative_width_eV",
    "atomic_subshell_occupancy",
    "xray_transition",
    "xray_transition_notation",
    "xray_transition_energy_eV",
    "xray_transition_probability",
    "xray_transition_relative_weight",
    "xray_line",
]

# Standard library modules.
import os
import logging

# Third party modules.
import sqlalchemy

# Local modules.
from pyxray.base import _DatabaseMixin, NotFound
from pyxray.sql.data import SqlDatabase

# Globals and constants variables.
logger = logging.getLogger(__name__)


class _EmptyDatabase(_DatabaseMixin):
    def get_preferred_references(self):
        return ()

    def add_preferred_reference(self, reference):
        pass

    def clear_preferred_references(self):
        pass

    def element(self, element):  # pragma: no cover
        raise NotFound

    def element_atomic_number(self, element):  # pragma: no cover
        raise NotFound

    def element_symbol(self, element, reference=None):  # pragma: no cover
        raise NotFound

    def element_name(self, element, language="en", reference=None):  # pragma: no cover
        raise NotFound

    def element_atomic_weight(self, element, reference=None):  # pragma: no cover
        raise NotFound

    def element_mass_density_kg_per_m3(
        self, element, reference=None
    ):  # pragma: no cover
        raise NotFound

    def element_mass_density_g_per_cm3(
        self, element, reference=None
    ):  # pragma: no cover
        raise NotFound

    def element_xray_transitions(
        self, element, xray_transition=None, reference=None
    ):  # pragma: no cover
        raise NotFound

    def element_xray_transition(self, element, xray_transition, reference=None):
        raise NotFound

    def atomic_shell(self, atomic_shell):  # pragma: no cover
        raise NotFound

    def atomic_shell_notation(
        self, atomic_shell, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        raise NotFound

    def atomic_subshell(self, atomic_subshell):  # pragma: no cover
        raise NotFound

    def atomic_subshell_notation(
        self, atomic_subshell, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        raise NotFound

    def atomic_subshell_binding_energy_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        raise NotFound

    def atomic_subshell_radiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        raise NotFound

    def atomic_subshell_nonradiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        raise NotFound

    def atomic_subshell_occupancy(
        self, element, atomic_subshell, reference=None
    ):  # pragma: no cover
        raise NotFound

    def xray_transition(self, xraytransition):  # pragma: no cover
        raise NotFound

    def xray_transition_notation(
        self, xray_transition, notation, encoding="utf16", reference=None
    ):  # pragma: no cover
        raise NotFound

    def xray_transition_energy_eV(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        raise NotFound

    def xray_transition_probability(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        raise NotFound

    def xray_transition_relative_weight(
        self, element, xray_transition, reference=None
    ):  # pragma: no cover
        raise NotFound

    def xray_line(self, element, xray_transition):  # pragma: no cover
        raise NotFound


def _init_sql_database():
    basedir = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(basedir, "data", "pyxray.db")
    if not os.path.exists(filepath):
        raise RuntimeError("Cannot find SQL database at location {0}".format(filepath))

    engine = sqlalchemy.create_engine("sqlite:///" + filepath)
    return SqlDatabase(engine)


try:
    database = _init_sql_database()
except:
    logger.error("No SQL database found")
    database = _EmptyDatabase()

get_preferred_references = database.get_preferred_references
add_preferred_reference = database.add_preferred_reference
clear_preferred_references = database.clear_preferred_references
element = database.element
element_atomic_number = database.element_atomic_number
element_symbol = database.element_symbol
element_name = database.element_name
element_atomic_weight = database.element_atomic_weight
element_mass_density_kg_per_m3 = database.element_mass_density_kg_per_m3
element_mass_density_g_per_cm3 = database.element_mass_density_g_per_cm3
element_xray_transitions = database.element_xray_transitions
element_xray_transition = database.element_xray_transition
print_element_xray_transitions = database.print_element_xray_transitions
atomic_shell = database.atomic_shell
atomic_shell_notation = database.atomic_shell_notation
atomic_subshell = database.atomic_subshell
atomic_subshell_notation = database.atomic_subshell_notation
atomic_subshell_binding_energy_eV = database.atomic_subshell_binding_energy_eV
atomic_subshell_radiative_width_eV = database.atomic_subshell_radiative_width_eV
atomic_subshell_nonradiative_width_eV = database.atomic_subshell_nonradiative_width_eV
atomic_subshell_occupancy = database.atomic_subshell_occupancy
xray_transition = database.xray_transition
xray_transition_notation = database.xray_transition_notation
xray_transition_energy_eV = database.xray_transition_energy_eV
xray_transition_probability = database.xray_transition_probability
xray_transition_relative_weight = database.xray_transition_relative_weight
xray_line = database.xray_line



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.base import _DatabaseMixin

# Globals and constants variables.


class MockDatabase(_DatabaseMixin):
    def element(self, element):
        pass

    def element_atomic_number(self, element):
        pass

    def element_symbol(self, element, reference=None):
        pass

    def element_name(self, element, language="en", reference=None):
        pass

    def element_atomic_weight(self, element, reference=None):
        pass

    def element_mass_density_kg_per_m3(self, element, reference=None):
        pass

    def element_mass_density_g_per_cm3(self, element, reference=None):
        pass

    def element_xray_transitions(self, element, reference=None):
        pass

    def element_xray_transition(self, element, xraytransitionset=None, reference=None):
        pass

    def atomic_shell(self, atomic_shell):
        pass

    def atomic_shell_notation(
        self, atomic_shell, notation, encoding="utf16", reference=None
    ):
        pass

    def atomic_subshell(self, atomic_subshell):
        pass

    def atomic_subshell_notation(
        self, atomic_subshell, notation, encoding="utf16", reference=None
    ):
        pass

    def atomic_subshell_binding_energy_eV(
        self, element, atomic_subshell, reference=None
    ):
        pass

    def atomic_subshell_radiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):
        pass

    def atomic_subshell_nonradiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):
        pass

    def atomic_subshell_occupancy(self, element, atomic_subshell, reference=None):
        pass

    def xray_transition(self, xraytransition):
        pass

    def xray_transition_notation(
        self, xraytransition, notation, encoding="utf16", reference=None
    ):
        pass

    def xray_transition_energy_eV(self, element, xraytransition, reference=None):
        pass

    def xray_transition_probability(self, element, xraytransition, reference=None):
        pass

    def xray_transition_relative_weight(self, element, xraytransition, reference=None):
        pass

    def xray_transitionset(self, xraytransitionset):
        pass

    def xray_transitionset_notation(
        self, xraytransitionset, notation, encoding="utf16", reference=None
    ):
        pass

    def xray_transitionset_energy_eV(self, element, xraytransitionset, reference=None):
        pass

    def xray_transitionset_relative_weight(
        self, element, xraytransitionset, reference=None
    ):
        pass

    def xray_line(self, element, line, reference=None):
        pass


@pytest.fixture
def database():
    return MockDatabase()


# def test_base_get_default_reference(database):
#     assert database.get_default_reference('element_symbol') is None

#     with pytest.raises(ValueError):
#         database.get_default_reference('foo')

# def test_base_set_default_reference(database):
#     database.set_default_reference('element_symbol', 'doe2016')

#     assert database.get_default_reference('element_symbol') == 'doe2016'

#     with pytest.raises(ValueError):
#         database.set_default_reference('foo', 'doe2016')



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.parser.campbell2001 import CampbellAtomicSubshellRadiativeWidthParser

# Globals and constants variables.


def test_campbell2011():
    parser = CampbellAtomicSubshellRadiativeWidthParser()

    props = list(parser)
    assert len(props) == 963

    assert props[0].element.z == 10
    assert props[0].value_eV == pytest.approx(0.24, abs=1e-2)

    assert props[962].element.z == 92
    assert props[962].value_eV == pytest.approx(0.31, abs=1e-2)
